/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include <dirent.h>
#include <mntent.h>
#include <stdlib.h>
#include <linux/hdreg.h>
#include <sys/ioctl.h>
#include "internal.h"
#include "ata.h"
#include "scsi.h"

/*
 * subdirectory below sysfs moint point holding the 
 * subdirectory hierarchies of all block devices.
 */
#define	BLOCK 		"/block"

/* Find sysfs mount point */
#ifndef	_PATH_MOUNTS
#define	_PATH_MOUNTS	"/proc/mounts"
#endif

static char *find_sysfs_mp(struct lib_context *lc)
{
	char *ret = NULL;
	FILE *mfile;
	struct mntent *ment;

	/* Try /proc/mounts first and failback to /etc/mtab. */
	if (!(mfile = setmntent(_PATH_MOUNTS, "r"))) {
		if (!(mfile = setmntent(_PATH_MOUNTED, "r")))
			LOG_ERR(lc, NULL, "Unable to open %s or %s",
				_PATH_MOUNTS, _PATH_MOUNTED);
	}

	while ((ment = getmntent(mfile))) {
		if (!strcmp(ment->mnt_type, "sysfs")) {
			ret = ment->mnt_dir;
			break;
		}
	};

	endmntent(mfile);

	return ret;
}

/* Make up an absolute sysfs path given a relative one */
static char *mk_sysfs_path(struct lib_context *lc, char const *path) {
	char *ret = NULL;
	char *sysfs_mp;

	if (!(sysfs_mp = find_sysfs_mp(lc)))
		LOG_ERR(lc, NULL, "finding sysfs mount point");

	if ((ret = dbg_malloc(strlen(sysfs_mp) + strlen(path) + 1)))
		sprintf(ret, "%s%s", sysfs_mp, path);
	else
		log_string_err(lc, __func__);

	return ret;
}

/* REMOVEME: Test with dm devices ? */
#ifdef	DMRAID_TEST
static int dm_device(struct lib_context *lc, char *path)
{
	return !strncmp(get_basename(lc, path), "dm-", 3);
}

/* Fake a SCSI serial number by reading it from a file. */
static int get_dm_serial(struct lib_context *lc,
			 struct dev_info *di, char *path)
{
#define	DOT_SERIAL	".serial"
	int ret = 0;
	FILE *f;
	char *serial, buffer[64];

	if (!(serial = dbg_malloc(strlen(path) + sizeof(DOT_SERIAL) + 1)))
		return log_string_err(lc, __func__);

	sprintf(serial, "%s%s", path, DOT_SERIAL);
	if ((f = fopen(serial, "r")) &&
	    fgets(buffer, 32, f) &&
	    (di->serial = dbg_strdup(remove_space(lc, buffer, strlen(buffer)))))
		ret = 1;

	dbg_free(serial);
	if (f)
		fclose(f);
	else
		log_warn(lc, "missing dm serial file for %s", di->path);

	return ret;
#undef	DOT_SERIAL
}
#endif

/*
 * Ioctl for sector, optionally for device size
 * and get device serial number.
 */
static int get_device_serial(struct lib_context *lc, int fd,
			     struct dev_info *di)
{
	/*
	 * In case new generic SCSI ioctl fails,
	 * try ATA and fall back to old SCSI ioctl.
	 */
	return get_scsi_serial(lc, fd, di, SG) || /* SG: generic scsi ioctl. */
	       get_ata_serial(lc, fd, di) ||
	       get_scsi_serial(lc, fd, di, OLD);  /* OLD: Old scsi ioctl. */
}

static int di_ioctl(struct lib_context *lc, int fd, struct dev_info *di)
{
	unsigned int sector_size = 0;
	unsigned long size;

	/* Fetch sector size. */
	if (ioctl(fd, BLKSSZGET, &sector_size))
		sector_size = DMRAID_SECTOR_SIZE;

	if (sector_size != DMRAID_SECTOR_SIZE)
		LOG_ERR(lc, 0, "unsupported sector size %d on %s.",
			sector_size, di->path);

	/* Use size device ioctl in case we didn't get the size from sysfs. */
	if (!di->sectors && !ioctl(fd, BLKGETSIZE, &size))
		di->sectors = size;

	/* REMOVEME: Test with dm devices ? */
#ifdef	DMRAID_TEST
	if (dm_device(lc, di->path))
		get_dm_serial(lc, di, di->path);
	else {
#endif
		if (!get_device_serial(lc, fd, di))
			return 0;
#ifdef	DMRAID_TEST
	}
#endif

	return 1;
}

/* Are we interested in this device ? */
static int interested(struct lib_context *lc, char *name)
{
	/*
	 * Whole IDE and SCSI disks only.
	 * Include dm devices for testing.
	 */
	return ((!isdigit(name[strlen(name)-1]) &&
	        *(name+1) == 'd' &&
	        (*name == 'h' || *name == 's'))
#ifdef	DMRAID_TEST
		|| dm_device(lc, name)
#endif
	       ) ? 1 : 0;
}

/* Ask sysfs, if a device is removable. */
int removable_device(struct lib_context *lc, char *dev_path)
{
#define	SYSFS_REMOVABLE	"removable"
	int ret = 0;
	char *name = get_basename(lc, dev_path), *sysfs_path, *sysfs_file;
	FILE *f;

	if (!(sysfs_path = mk_sysfs_path(lc, BLOCK)))
		return 0;

	if (!(sysfs_file = dbg_malloc(strlen(sysfs_path) + strlen(name) +
				      sizeof(SYSFS_REMOVABLE) + 3))) {
		log_string_err(lc, __func__);
		goto out;
	}

	sprintf(sysfs_file, "%s/%s/%s", sysfs_path, name, SYSFS_REMOVABLE);
	if ((f = fopen(sysfs_file, "r"))) {
		unsigned short removable = 0;

		if (fscanf(f, "%hu", &removable) == 1 &&
		    removable) {
			log_notice(lc, "skipping removable device %s",
				   dev_path);
			ret = 1;
		}

		fclose(f);
	}

	dbg_free(sysfs_file);

   out:
	dbg_free(sysfs_path);

	return ret;
#undef	SYSFS_REMOVABLE
}

/*
 * Read the size in sectors from the sysfs "size" file.
 * Avoid access to removable devices.
 */
static int sysfs_get_size(struct lib_context *lc, struct dev_info *di,
			  const char *path, char *name)
{
#define	SYSFS_SIZE	"size"
	int ret = 0;
	char *sysfs_file;
	FILE *f;

	if (!(sysfs_file = dbg_malloc(strlen(path) + strlen(name) +
				      sizeof(SYSFS_SIZE) + 3)))
		return log_string_err(lc, __func__);

	sprintf(sysfs_file, "%s/%s/%s", path, name, SYSFS_SIZE);
	if ((f = fopen(sysfs_file, "r"))) {
		if ((ret = fscanf(f, "%" PRIu64, &di->sectors)) != 1) {
			ret = 0;
			log_err(lc, "reading disk size for %s from sysfs",
				di->path);
		}

		fclose(f);
	} else
		log_err(lc, "opening %s", sysfs_file);

	dbg_free(sysfs_file);

	return ret;
#undef	SYSFS_SIZE
}

static int get_size(struct lib_context *lc, char *path,
		    char *name, int sysfs)
{
	int fd, ret = 0;
	char *dev_path;
	struct dev_info *di = NULL;

	if (!interested(lc, name))
		return 1;

	if (!(dev_path = dbg_malloc(strlen(_PATH_DEV) + strlen(name) + 1)))
		return log_string_err(lc, __func__);

	sprintf(dev_path, "%s%s", _PATH_DEV, name);
	if (removable_device(lc, dev_path))
		goto out;

	if (!(di = alloc_dev_info(lc, dev_path)))
		goto out;

	if (sysfs && !sysfs_get_size(lc, di, path, name))
		goto out;

	if ((fd = open(dev_path, O_RDONLY)) == -1)
		goto out;

	if (!di_ioctl(lc, fd, di))
		goto close;

	list_add(&di->list, LC_DI(lc));
	ret = 1;

   close:
	close(fd);
   out:
	dbg_free(dev_path);

	if (!ret && di)
		free_dev_info(lc, di);

	return ret;
}

/*
 * Find disk devices in sysfs or directly
 * in /dev on Linux 2.4 and keep information.
 */
int discover_devices(struct lib_context *lc)
{
	int sysfs, ret = 0;
	char *path, *p;
	DIR *d;
	struct dirent *de;

	if ((p = mk_sysfs_path(lc, BLOCK))) {
		path = p;
		sysfs = 1;
	} else {
		path = (char*) _PATH_DEV;
		sysfs = 0;
		log_print(lc, "carrying on with %s", path);
	}

	if (!(d = opendir(path))) {
		log_err(lc, "opening path %s", path);
		goto out;
	}

	while ((de = readdir(d)))
		get_size(lc, path, de->d_name, sysfs);

	closedir(d);
	ret = 1;

  out:
	if (p)
		dbg_free(p);

	return ret;
}
