/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include "internal.h"

#define	ARRAY_LIMIT(array, idx) \
	((idx) < ARRAY_SIZE(array) ? (idx) : ARRAY_SIZE(array) - 1)

/* Display information about a block device */
static void log_disk(struct lib_context *lc, struct dev_info *di)
{
	const char *fmt[] = {
		"%s: %12" PRIu64 " total, \"%s\"",
		"%s",
		"%s:%" PRIu64 ":\"%s\"",
	};

	log_print(lc, fmt[ARRAY_LIMIT(fmt, lc_opt(lc, LC_COLUMN))],
		  di->path, di->sectors, di->serial ? di->serial : "N/A");
}

/* Turn NULL (= "unknown") into a displayable string. */
static const char *check_null(const char *str)
{
	return str ? str : "unknown";
}

/* Display information about a RAID device */
static void log_rd(struct lib_context *lc, struct raid_dev *rd)
{
	const char *fmt[] = {
		"%s: %s, \"%s\", %s, %s, %" PRIu64 " sectors, data@ %" PRIu64,
		"%s",
		"%s:%s:%s:%s:%s:%" PRIu64 ":%" PRIu64,
	};

	log_print(lc, fmt[ARRAY_LIMIT(fmt, lc_opt(lc, LC_COLUMN))],
		  rd->di->path, rd->fmt->name, rd->name,
		  check_null(get_type(lc, rd->type)),
		  check_null(get_status(lc, rd->status)),
		  rd->sectors, rd->offset);
}

/* Log native RAID device information. */
static void log_rd_native(struct lib_context *lc, struct raid_dev *rd)
{
	if (rd->fmt->log) {
		rd->fmt->log(lc, rd);
		log_print(lc, " ");
	} else
		log_print(lc, "\"%s\" doesn't support native logging of RAID "
			  "device information", rd->fmt->name);
}

/* Dispatch log functions. */
static void log_devices(struct lib_context *lc, enum dev_type type)
{
	struct dev_info *di;
	struct raid_dev *rd;

	switch (type) {
	case DEVICE:
		list_for_each_entry(di, LC_DI(lc), list)
			log_disk(lc, di);
		break;

	case NATIVE:
		list_for_each_entry(rd, LC_RD(lc), list)
			log_rd_native(lc, rd);
		break;

	case RAID:
		list_for_each_entry(rd, LC_RD(lc), list)
			log_rd(lc, rd);
		break;

	default:
		log_err(lc, "%s: unknown device type", __func__);
	}
}

/* Display information about a dmraid format handler */
static void log_format(struct lib_context *lc, struct dmraid_format *fmt)
{
	log_print(lc, "%-6s : %s", fmt->name, fmt->descr);
}

/* Display information about devices depending on device type. */
void display_devices(struct lib_context *lc, enum dev_type type)
{
	int devs;

	if ((devs = count_devices(lc, type))) {
		log_info(lc, "%s device%s discovered:\n",
			 (type & (RAID|NATIVE)) ? "RAID" : "Block",
			  devs == 1 ? "" : "s");

		log_devices(lc, type);
	}
}

static void _display_set(struct lib_context *lc, struct raid_set *rs)
{
	unsigned int devs = 0, o, spares = 0, subsets = 0;
	uint64_t sectors = 0;
	const char *fmt[] = {
		"name   : %s\n"
		"size   : %" PRIu64 "\n"
		"stride : %u\n"
		"type   : %s\n"
		"status : %s\n"
		"subsets: %u\n"
		"devs   : %u\n"
		"spares : %u",
		"%s",
		"%s:%" PRIu64 ":%u:%s:%s:%u:%u:%u",
	};

	if (T_GROUP(rs) && !lc_opt(lc, LC_GROUP))
		return;

	if ((o = ARRAY_LIMIT(fmt, lc_opt(lc, LC_COLUMN))) != 1) {
		sectors = total_sectors(lc, rs);
		subsets = count_sets(lc, &rs->sets);
		devs    = count_devs(lc, rs, ct_dev);
		spares  = count_devs(lc, rs, ct_spare);
	}

	log_print(lc, fmt[o],
		  rs->name, sectors, rs->stride,
		  check_null(get_set_type(lc, rs)),
		  check_null(get_status(lc, rs->status)),
		  subsets, devs, spares);

	if (lc_opt(lc, LC_COLUMN) > 2) {
		struct raid_dev *rd;

		list_for_each_entry(rd, &rs->devs, devs)
			log_rd(lc, rd);
	}
}

/* FIXME: Spock, do something better (neater). */
static int group_active(struct lib_context *lc, struct raid_set *rs)
{
	struct raid_set *r;

	list_for_each_entry(r, &rs->sets, list) {
		if (dm_status(lc, r))
			return 1;
	}

	return 0;
}

void display_set(struct lib_context *lc, void *v,
		 enum active_type active, int top)
{
	struct raid_set *rs = v;
	struct raid_set *r;
	int dmstatus = T_GROUP(rs) ? group_active(lc, rs) : dm_status(lc, rs);

	if (((active & A) && !dmstatus) ||
	   ((active & I) && dmstatus))
		return;

	if (!lc_opt(lc, LC_COLUMN)) {
		if (T_GROUP(rs) && !lc_opt(lc, LC_GROUP))
			log_print(lc, "*** Group superset %s", rs->name);
		else {
			log_print(lc, "%s %s%s%set",
		  		top ? "-->" : "***",
		  		S_INCONSISTENT(rs->status) ?
				"*Inconsistent* " : "",
		  		dm_status(lc, rs) ? "Active " : "",
		  		SETS(rs) ? "Supers" : (top ? "Subs" : "S"));
		}
	}

	_display_set(lc, rs);

	/* Optionally display subsets. */
	if (T_GROUP(rs) || /* Always display for GROUP sets. */
	    lc_opt(lc, LC_SETS) > 1 ||
	    lc_opt(lc, LC_COLUMN) > 2) {
		list_for_each_entry(r, &rs->sets, list)
			display_set(lc, r, active, ++top);
	}
}

/*
 * Display information about supported RAID metadata formats
 * (ie. registered format handlers)
 */
static void _list_formats(struct lib_context *lc, enum fmt_type type)
{
	struct format_list *fmt_list;

	list_for_each_entry(fmt_list, LC_FMT(lc), list) {
		if (type == fmt_list->fmt->format)
			log_format(lc, fmt_list->fmt);
	}
}

int list_formats(struct lib_context *lc, int arg)
{
	if (!list_empty(LC_FMT(lc)))
		log_notice(lc, "supported metadata formats:\n");

	_list_formats(lc, FMT_RAID);
	_list_formats(lc, FMT_PARTITION);

	log_print(lc, "(0): Discover, (+): Discover+Activate");
	return 1;
}
