/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Intel Software RAID metadata format handler.
 *
 * isw_read() profited from Carl-Daniel Hailfinger's raiddetect code.
 *
 * Profited from the Linux 2.4 iswraid driver by
 * Boji Tony Kannanthanam and Martins Krikis.
 */

#define	HANDLER	"isw"

#include "internal.h"
#define	FORMAT_HANDLER
#include "isw.h"
#include <datastruct/byteorder.h>

static const char *handler = HANDLER;

/*
 * Make up RAID set name from family_num and volume name.
 */
static size_t _name(struct isw *isw, struct isw_dev *dev,
		     char *str, size_t len)
{
	return snprintf(str, len, "isw_%u%s%s", isw->family_num,
			dev ? "_" : "", dev ? (char*) dev->volume : "") + 1;
}

static char *name(struct lib_context *lc, struct isw *isw, struct isw_dev *dev)
{
        size_t len;
        char *ret;

        if ((ret = dbg_malloc((len = _name(isw, dev, NULL, 0)) + 1))) {
                _name(isw, dev, ret, len);
		mk_alpha(lc, ret + HANDLER_LEN, len - HANDLER_LEN -
			 (dev ? strlen((char*) dev->volume) - 2 : 1));
        } else
		log_string_err(lc, handler);

        return ret;
}

/* Find a disk table slot by serial number. */
static struct isw_disk *get_disk(struct lib_context *lc,
				 struct isw *isw, struct dev_info *di)
{
	unsigned int i;

	if (!di->serial)
		LOG_ERR(lc, NULL, "%s: no SCSI serial number", handler);

	for (i = 0; i < isw->num_disks; i++) {
		if (!strcmp(di->serial, isw->disk[i].serial))
			return &isw->disk[i];
	}

	return NULL;
}

/*
 * Retrieve status of device.
 *
 * FIXME: is this sufficient to cover all state ?
 */
static enum status __status(unsigned int status)
{
	return ((CONFIGURED_DISK & status) && (USABLE_DISK & status)) ?
	       s_ok : s_broken;
}

static enum status status(struct lib_context *lc, struct raid_dev *rd)
{
	struct isw *isw = META(rd, isw);
	struct isw_disk *disk = get_disk(lc, isw, rd->di);

	if (disk)
		return __status(disk->status);

	LOG_ERR(lc, s_undef, "%s: Error finding disk table slot for %s",
		handler, rd->di->path);
}

/*
 * Mapping of Intel types to generic types.
 */
static struct types types[] = {
        { ISW_T_RAID0, t_raid0},
        { ISW_T_RAID1, t_raid1},
        { 0, t_undef},
};

/* Neutralize disk type. */
static enum type type(struct raid_dev *rd)
{
	struct isw_dev *dev = rd->private.ptr;

	if (dev)
		return rd_type(types, (unsigned int) dev->vol.map.raid_level);

	return t_group;
}

/*
 * Generate checksum of Raid metadata for mpb_size/sizeof(u32) words
 * (checksum field itself ignored for this calculation).
 */
static uint32_t _checksum(struct isw *isw)
{
	uint32_t end = isw->mpb_size / sizeof(end), ret = 0;
	uint32_t *p = (uint32_t*) isw;

	while (end--)
		ret += *p++;

	return ret - isw->check_sum;
}

/* Advance to the next isw_dev from a given one. */
static struct isw_dev *advance_raiddev(struct isw_dev *dev)
{
	int i; /* map size correction */
	struct isw_vol *vol = &dev->vol;
	struct isw_map *map = &vol->map;

	/* Correction */
	i = (map->num_members - 1) * sizeof(map->disk_ord_tbl);
	dev++; /* off by i bytes for sure */
	dev = (struct isw_dev*) ((uint8_t*) dev + i);

	if (vol->migr_state) { /* need to add space for another map */
		map = (struct isw_map*) dev; /* Yes, it sits here! */
		i = (map->num_members - 1) * sizeof(map->disk_ord_tbl);
		dev = (struct isw_dev*) ((uint8_t*) dev + sizeof(*map) + i);
	}

	return dev;
}

/* Return isw_dev by table index. */
static struct isw_dev *raiddev(struct isw *isw, unsigned int i)
{
	struct isw_dev *dev = (struct isw_dev*) &isw->disk[isw->num_disks];

	while(i--)
		dev = advance_raiddev(dev);

	return dev;
}

/*
 * Read an Intel RAID device
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x, y)
#else
static void to_cpu(struct isw *isw, int full)
{
	unsigned int i, j;
	struct isw_disk *disk;
	struct isw_dev *dev;

	CVT32(isw->check_sum);
	CVT32(isw->mpb_size);
	CVT32(isw->family_num);
	CVT32(isw->generation_num);

	if (!full)
		return;

	for (i = 0; i < isw->num_disks; i++) {
		disk = &isw->disk[i];

		CVT32(disk->totalBlocks);
		CVT32(disk->scsiId);
		CVT32(disk->status);
	}

	for (i = 0; i < isw->num_raid_devs; i++) {
		dev = raiddev(isw, i);

		/* RAID device. */
		CVT32(dev->SizeLow);
		CVT32(dev->SizeHigh);
		CVT32(dev->status);
		CVT32(dev->reserved_blocks);

		/* RAID volume only has 8 bit members. */
	
		/* RAID map. */
		CVT32(dev->vol.map.pba_of_lba0);
		CVT32(dev->vol.map.blocks_per_member);
		CVT32(dev->vol.map.num_data_stripes);
		CVT16(dev->vol.map.blocks_per_strip);

		for (j = 0; j < dev->vol.map.num_members; j++)
			CVT16(dev->vol.map.disk_ord_tbl[j]);
	}
}
#endif

static int is_isw(struct lib_context *lc, struct dev_info *di, struct isw *isw)
{
	if (strncmp(isw->sig, MPB_SIGNATURE, sizeof(MPB_SIGNATURE) - 1))
		return 0;

	/* Check version info, older versions supported */
	if (strncmp(isw->sig + sizeof(MPB_SIGNATURE) - 1,
		   MPB_VERSION_RAID2, sizeof(MPB_VERSION_RAID2) - 1) > 0) {
		log_print(lc, "%s: superblock found on %s, "
			  "but version is too high", handler, di->path);
		return 0;
	}

	return 1;
}

static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct isw *isw, struct dev_info *di, loff_t isw_sboffset);
static struct raid_dev *isw_read(struct lib_context *lc, struct dev_info *di)
{
	size_t size = ISW_DISK_BLOCK_SIZE;
	loff_t mpbblocks, isw_sboffset = ISW_CONFIGOFFSET;
	struct isw *isw, *isw_big;
	struct raid_dev *rd = NULL;

	if (!(isw = alloc_private_and_read(lc, handler, size,
					   di->path, isw_sboffset)))
		goto out;

	/* Check Signature */
	to_cpu(isw, 0);
	if (!is_isw(lc, di, isw))
		goto bad;

	/* Get the rounded up value for mpbblocks */
	if ((mpbblocks = ((isw->mpb_size + ISW_DISK_BLOCK_SIZE - 1) /
			  ISW_DISK_BLOCK_SIZE)) > 1) {
		size = mpbblocks * ISW_DISK_BLOCK_SIZE;

		/* Allocate memory for extend Intel superblock */
		if (!(isw_big = alloc_private(lc, handler, size)))
			goto bad;
	
		/* Save the first metadata sector */
		memcpy(isw_big, isw, ISW_DISK_BLOCK_SIZE);
		dbg_free(isw);
		isw = isw_big;
	
		/*
		 * Calculate offset to extended Intel superblock
		 * and try to read it.
		 */
		isw_sboffset -= (mpbblocks - 1) * ISW_DISK_BLOCK_SIZE;
		if (!read_file(lc, handler, di->path,
			       (void*) isw + ISW_DISK_BLOCK_SIZE,
			       size - ISW_DISK_BLOCK_SIZE, isw_sboffset)) {
			log_err(lc, "%s: reading extended "
				"superblock for %s", handler, di->path);
			goto bad;
		}
	}

	/* Calculate superblock checksum */
	to_cpu(isw, 1);
	file_metadata(lc, handler, di->path, isw, size, isw_sboffset >> 9);

	if (isw->check_sum != _checksum(isw)) {
		log_err(lc, "%s: extended superblock for %s has wrong checksum",
			handler, di->path);
			goto bad;
	}

	if (!(rd = alloc_raid_dev(lc, handler)))
		goto bad;

	if (setup_rd(lc, rd, isw, di, isw_sboffset))
		goto out;

	log_err(lc, "%s: setting up RAID device %s", handler, di->path);
	free_raid_dev(lc, &rd);
	goto out;

   bad:
	dbg_free(isw);
   out:
	return rd;
}

/*
 * Write an Intel Software RAID device.
 */
static int isw_write(struct lib_context *lc, struct raid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct isw *isw = META(rd, isw);
#endif

	to_disk(isw, 1);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(isw, 1);
	return ret;
}

/*
 * Group an Intel SW RAID disk into potentially
 * multiple RAID sets and RAID disks.
 */
/* Create a RAID device to map a volumes segment. */
static struct raid_dev *_create_rd(struct lib_context *lc,
				     struct raid_dev *rd,
				     struct isw *isw, struct isw_dev *dev)
{
	struct raid_dev *r;

	/* FIXME: FAILED_MAP etc. */
	if (dev->vol.map.map_state) {
		log_print(lc, "%s: unsupported map state 0x%x on %s for %s",
			  handler, dev->vol.map.map_state, rd->di->path,
			  (char*) dev->volume);
		return NULL;
	}

	if (!(r = alloc_raid_dev(lc, handler)))
		return NULL;

	if ((r->private.ptr = alloc_private(lc, handler, sizeof*dev)))
		memcpy(r->private.ptr, dev, sizeof(*dev));
	else {
		log_err(lc, "%s: alloctaing private memory", handler);
		goto free;
	}

	if (type(r) == t_undef) {
		log_err(lc, "%s: RAID type %u not supported",
			handler, (unsigned int) dev->vol.map.raid_level);
		goto free;
	}

	r->di = rd->di;
        if (!(r->name = name(lc, isw, dev)))
		goto free;

	r->fmt = rd->fmt;
	
	r->offset = dev->vol.map.pba_of_lba0;
	r->sectors = dev->vol.map.blocks_per_member;

	r->type   = type(r);
	goto out;

   free:
	free_raid_dev(lc, &r);
   out:
	return r;
}

/* Find an Intel RAID set or create it. */
static void create_rs(struct raid_set *rs, void* private)
{
	if (private)
		rs->stride =
			((struct isw_dev*) private)->vol.map.blocks_per_strip;
}

/*
 * rs_group contains the top-level group RAID set (type: t_group) on entry
 * and shall be returned on success (or NULL on error).
 */
static struct raid_set *group_rd(struct lib_context *lc,
				   struct raid_set *rs_group,
				   struct raid_dev *rd_meta)
{
	unsigned int d;
	struct isw *isw = META(rd_meta, isw);
	struct isw_dev *dev;
	struct raid_dev *rd;
	struct raid_set *rs;

	/* Loop the device/volume table. */
	for (d = 0; d < isw->num_raid_devs; d++) {
		dev = raiddev(isw, d);

		/* FIXME: sufficient ? */
		if (dev->status) {
			log_err(lc, "%s: volume %s is broken",
				handler, dev->volume);
			continue;
		}

		if (!(rd = _create_rd(lc, rd_meta, isw, dev)))
			return NULL;

		if (!(rs = find_or_alloc_raid_set(lc, rd->name, FIND_ALL,
				      		  rd, &rs_group->sets,
						  create_rs, dev))) {
			free_raid_dev(lc, &rd);
			return NULL;
		}

		list_add_tail(&rd->devs, &rs->devs);
	}

	return rs_group;
}

/* Add an Intel SW RAID device to a set */
static struct raid_set *isw_group(struct lib_context *lc,
				    struct raid_dev *rd_meta)
{
	struct raid_set *rs_group;

	/*
	 * Once we get here, an Intel SW RAID disk containing a metadata area
	 * with a volume table has been discovered by isw_read.
	 */
	/* Check if a top level group RAID set already exists. */
	if (!(rs_group = find_or_alloc_raid_set(lc, rd_meta->name, FIND_TOP,
				      		rd_meta, LC_RS(lc),
						NO_CREATE, NO_CREATE_ARG)))
		return NULL;

	/* Add the whole underlying (meta) RAID device to the group set. */
	list_add_tail(&rd_meta->devs, &rs_group->devs);

	/*
	 * We need to run through the volume table and create a RAID set and
	 * RAID devices hanging off it for every volume,
	 * so that the activate code is happy.
	 *
	 * A pointer to the top-level group RAID set
	 * gets returned or NULL on error.
	 */
	return group_rd(lc, rs_group, rd_meta);
}

/*
 * Check an Intel SW RAID set.
 *
 * FIXME: more sanity checks.
 */
static void __isw_check(struct lib_context *lc, struct raid_set *rs)
{
	struct raid_dev *rd;

	if (T_GROUP(rs))
		return;

	list_for_each_entry(rd, &rs->devs, devs) {
		struct isw_dev *dev = (struct isw_dev*) rd->private.ptr;

		/* FIXME: more status checks ? */
		rd->status = dev->status ? s_broken : s_ok;
		if (S_BROKEN(rd->status))
			log_err(lc, "%s device for volume \"%s\" broken on %s "
				"in RAID set \"%s\"",
				handler, dev->volume, rd->di->path, rs->name);

		log_dbg(lc, "set status of %s device on \"%s\" to %u", handler,
			rd->di->path, rd->status);
	}
}

static int _isw_check(struct lib_context *lc, struct raid_set *rs)
{
	/* Check any hierachical RAID sets. */
	check_set(lc, rs, _isw_check, handler);

	__isw_check(lc, rs);

	return set_rs_status(lc, rs);
}

static int isw_check(struct lib_context *lc, struct raid_set *rs_group)
{
	return T_GROUP(rs_group) ? _isw_check(lc, rs_group) : 0;
}

/*
 * IO error event handler.
 */
static int event_io(struct lib_context *lc, struct event_io *e_io)
{
	struct raid_dev *rd = e_io->rd;
	struct isw *isw = META(rd, isw);
	struct isw_disk *disk;

	if (!(disk = get_disk(lc, isw, rd->di)))
		LOG_ERR(lc, 0, "%s: disk", handler);

	/* Avoid write trashing. */
	if (S_BROKEN(status(lc, rd)))
		return 0;

	disk->status &= ~USABLE_DISK;

	return 1;
}

static struct event_handlers isw_event_handlers = {
	.io = event_io,
	.rd = NULL,	/* FIXME: no device add/remove event handler yet. */
};

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about an ISW RAID device.
 */
static void isw_log(struct lib_context *lc, struct raid_dev *rd)
{
	unsigned int d, i;
	struct isw *isw = META(rd, isw);
	struct isw_dev *dev;

	log_print(lc, "%s (%s):", rd->di->path, handler);
	P("sig: \"%*s\"", isw, isw->sig, MAX_SIGNATURE_LENGTH, isw->sig);
	DP("check_sum: %u", isw, isw->check_sum);
	DP("mpb_size: %u", isw, isw->mpb_size);
	DP("family_num: %u", isw, isw->family_num);
	DP("generation_num: %u", isw, isw->generation_num);
	DP("reserved[0]: %u", isw, isw->reserved[0]);
	DP("reserved[1]: %u", isw, isw->reserved[1]);
	DP("num_disks: %u", isw, isw->num_disks);
	DP("num_raid_devs: %u", isw, isw->num_raid_devs);
	DP("fill[0]: %u", isw, isw->fill[0]);
	DP("fill[1]: %u", isw, isw->fill[1]);
	for (i = 0; i < ISW_FILLERS; i++) {
		if (isw->filler[i])
        		P("filler[%i]: %u", isw,
			  isw->filler[i], i, isw->filler[i]);
	}

	/* Disk table. */
	for (d = 0; d < isw->num_disks; d++) {
		if (!isw->disk[d].totalBlocks)
			continue;

		P("disk[%u].serial: \"%*s\"", isw,
			  isw->disk[d].serial, d,
			  MAX_RAID_SERIAL_LEN, isw->disk[d].serial);
		P("disk[%u].totalBlocks: %u", isw,
			  isw->disk[d].totalBlocks,
			  d, isw->disk[d].totalBlocks);
		P("disk[%u].scsiId: 0x%x", isw,
			  isw->disk[d].scsiId, d, isw->disk[d].scsiId);
		P("disk[%u].status: 0x%x", isw,
			  isw->disk[d].status, d, isw->disk[d].status);
		for (i = 0; i < ISW_DISK_FILLERS; i++) {
			if (isw->disk[d].filler[i])
                		P("disk[%u].filler[%u]: %u", isw,
				  isw->disk[d].filler[i], d, i,
				  isw->disk[d].filler[i]);
		}
	}

	/* RAID device/volume table. */
	for (d = 0; d < isw->num_raid_devs; d++) {
		dev = raiddev(isw, d);

		/* RAID device */
		P("isw_dev[%u].volume: \"%*s\"", isw,
		  dev->volume, d, MAX_RAID_SERIAL_LEN, dev->volume);
		P("isw_dev[%u].SizeHigh: %u", isw,
		  dev->SizeHigh, d, dev->SizeHigh);
		P("isw_dev[%u].SizeLow: %u", isw,
		  dev->SizeLow, d, dev->SizeLow);
		P("isw_dev[%u].status: 0x%x", isw,
		  dev->status, d, dev->status);
		P("isw_dev[%u].reserved_blocks: %u", isw,
		  dev->reserved_blocks, d, dev->reserved_blocks);
		for (i = 0; i < ISW_DEV_FILLERS; i++) {
			if (dev->filler[i])
                		P("isw_dev[%u].filler[%u]: %u", isw,
				  dev->filler[i], d, i,
				  dev->filler[i]);
		}

		/* RAID volume */
		for (i = 0; i < 2; i++) {
			if (dev->vol.reserved[i])
                		P("isw_dev[%u].vol.reserved[%u]: %u", isw,
				  dev->vol.reserved[i], d, i,
				  dev->vol.reserved[i]);
		}
		P("isw_dev[%u].vol.migr_state: %u", isw,
		  dev->vol.migr_state, d, dev->vol.migr_state);
		P("isw_dev[%u].vol.migr_type: %u", isw,
		  dev->vol.migr_type, d, dev->vol.migr_type);
		P("isw_dev[%u].vol.dirty: %u", isw,
		  dev->vol.dirty, d, dev->vol.dirty);
		P("isw_dev[%u].vol.fill[0]: %u", isw,
		  dev->vol.fill[0], d, dev->vol.fill[0]);
		for (i = 0; i < 5; i++) {
			if (dev->vol.filler[i])
                		P("isw_dev[%u].vol.filler[%u]: %u", isw,
				  dev->vol.filler[i], d, i,
				  dev->vol.filler[i]);
		}

		/* RAID map */
		P("isw_dev[%u].vol.map.pba_of_lba0: %u", isw,
		  dev->vol.map.pba_of_lba0, d,
		  dev->vol.map.pba_of_lba0);
		P("isw_dev[%u].vol.map.blocks_per_member: %u", isw,
		  dev->vol.map.blocks_per_member, d,
		  dev->vol.map.blocks_per_member);
		P("isw_dev[%u].vol.map.num_data_stripes: %u", isw,
		  dev->vol.map.num_data_stripes, d,
		  dev->vol.map.num_data_stripes);
		P("isw_dev[%u].vol.map.blocks_per_strip: %u", isw,
		  dev->vol.map.blocks_per_strip, d,
		  dev->vol.map.blocks_per_strip);
		P("isw_dev[%u].vol.map.map_state: %u", isw,
		  dev->vol.map.map_state, d,
		  dev->vol.map.map_state);
		P("isw_dev[%u].vol.map.raid_level: %u", isw,
		  dev->vol.map.raid_level, d,
		  dev->vol.map.raid_level);
		P("isw_dev[%u].vol.map.num_members: %u", isw,
		  dev->vol.map.num_members, d,
		  dev->vol.map.num_members);
		for (i = 0; i < 3; i++) {
			if (dev->vol.map.reserved[i])
                		P("isw_dev[%u].vol.map.reserved[%u]: %u", isw,
				  dev->vol.map.reserved[i], d, i,
				  dev->vol.map.reserved[i]);
		}
		for (i = 0; i < 7; i++) {
			if (dev->vol.map.filler[i])
                		P("isw_dev[%u].vol.map.filler[%u]: %u", isw,
				  dev->vol.map.filler[i], d, i,
				  dev->vol.map.filler[i]);
		}
		for (i = 0; i < isw->num_disks; i++) {
			P("isw_dev[%u].vol.map.disk_ord_tbl[%u]: 0x%x", isw,
			  dev->vol.map.disk_ord_tbl[i], d, i,
			  dev->vol.map.disk_ord_tbl[i]);
		}
	}
}
#endif

static struct dmraid_format isw_format = {
	.name	= HANDLER,
	.descr	= "(+) Intel Software RAID",
	.format = FMT_RAID,
	.read	= isw_read,
	.write	= isw_write,
	.group	= isw_group,
	.check	= isw_check,
	.events	= &isw_event_handlers,
#ifdef DMRAID_NATIVE_LOG
	.log	= isw_log,
#endif
};

/* Register this format handler with the format core */
int register_isw(struct lib_context *lc)
{
	return register_format_handler(lc, &isw_format);
}

/*
 * Set the RAID device contents up derived from the Intel ones.
 *
 * This is the first one we get with here and we potentially need to
 * create many in isw_group() in case of multiple Intel SW RAID devices
 * on this RAID disk.
 */
static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct isw *isw, struct dev_info *di, loff_t isw_sboffset)
{
	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->di = di;
	rd->fmt = &isw_format;

	rd->offset = ISW_DATAOFFSET;
	rd->sectors = isw_sboffset / ISW_DISK_BLOCK_SIZE; /* FIXME: Correct ? */
	// rd->sectors -= rd->sectors & 1;

	rd->meta_areas->offset = rd->sectors;
	rd->meta_areas->size = di->sectors - rd->sectors - 1;
	rd->meta_areas->area = (void*) isw;

	rd->status = status(lc, rd);
	rd->type   = t_group;

        if ((rd->name = name(lc, isw, NULL)))
		return 1;

	return 0;
}
