/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * LSI Logic MegaRAID (and MegaIDE ?) ATARAID metadata format handler.
 *
 * Needs more metadata reengineering and grouping logic coding.
 *
 */

#define	HANDLER	"lsi"

#include "internal.h"
#define	FORMAT_HANDLER
#include "lsi.h"
#include <datastruct/byteorder.h>

static const char *handler = HANDLER;

/* Make up RAID device name. */
/* FIXME: senseful name ;) */
static size_t _name(struct lsi *lsi, char *str, size_t len)
{
	return snprintf(str, len, "lsi_%d", lsi->set_number) + 1;
}

static char *name(struct lib_context *lc, struct raid_dev *rd,
		  unsigned int subset)
{
	size_t len;
	char *ret;
	struct lsi *lsi = META(rd, lsi);

	len = _name(lsi, NULL, 0);
	if ((ret = dbg_malloc(len + 1))) {
		_name(lsi, ret, len);
		// mk_alpha(ret + HANDLER_LEN, strlen(ret) - HANDLER_LEN);
	} else
		log_string_err(lc, handler);

	return ret;
}

/* Mapping of LSI Logic types to generic types */
static struct types types[] = {
	{ LSI_T_RAID0, t_raid0 },
	{ LSI_T_RAID1, t_raid1 },
	{ LSI_T_RAID10, t_raid1 },
        { 0, t_undef}
};

/* Neutralize disk type */
static enum type type(struct lsi *lsi)
{
	return rd_type(types, (unsigned int) lsi->type);
}

/* LSI device status. */
/* FIXME: add flesh. */
static int status(struct lsi *lsi)
{
	return s_ok;
}

/* Decide about ordering sequence of RAID device. */
static int dev_sort(struct list_head *pos, struct list_head *new)
{
	return (META(RD(new), lsi))->disk_number <
	       (META(RD(pos), lsi))->disk_number;
}

/* Decide about ordering sequence of RAID set. */
static int set_sort(struct list_head *pos, struct list_head *new)
{
	return (META(RD_RS(RS(new)), lsi))->disk_number <
	       (META(RD_RS(RS(pos)), lsi))->disk_number;
}

/*
 * Read an LSI Logic RAID device
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct lsi *lsi)
{
	struct lsi_disks *disk;

	CVT16(lsi->stride);

	for (disk = lsi->disks; disk < &lsi->disks[4]; disk++) {
		CVT16(disk->magic_0);
		CVT16(disk->magic_1);
	}

	CVT32(lsi->what);
}
#endif

static int is_lsi(struct lsi *lsi)
{
	return !strncmp(lsi->magic_name, LSI_MAGIC_NAME,
			sizeof(LSI_MAGIC_NAME) - 1);
}

static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct lsi *lsi, struct dev_info *di);
static struct raid_dev *lsi_read(struct lib_context *lc, struct dev_info *di)
{
	struct lsi *lsi;
	struct raid_dev *rd = NULL;

	if (!(lsi = alloc_private_and_read(lc, handler, sizeof(*lsi),
					   di->path, LSI_CONFIGOFFSET)))
		goto out;

	to_cpu(lsi);
	if (!is_lsi(lsi))
		goto bad;

	file_metadata(lc, handler, di->path, lsi, sizeof(*lsi),
		      LSI_CONFIGOFFSET);

	if (!(rd = alloc_raid_dev(lc, handler)))
		goto bad;

	if (setup_rd(lc, rd, lsi, di))
		goto out;

	log_err(lc, "%s: setting up RAID device %s", handler, di->path);
	free_raid_dev(lc, &rd);
	goto out;

   bad:
	dbg_free(lsi);
   out:
	return rd;
}

/*
 * Write a LSI Logic RAID device.
 */
static int lsi_write(struct lib_context *lc, struct raid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct lsi *lsi = META(rd, lsi);
#endif

	to_disk(lsi);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(lsi);
	return ret;
}

/*
 * Group the RAID disk into a set.
 *
 * FIXME: this needs more work together with the metadata reengineering.
 */

static void super_created(struct raid_set *super, void *private)
{
	super->type = t_raid1;
	super->status = META((struct raid_dev *) private, lsi)->stride;
}

static int group_rd(struct lib_context *lc, struct raid_set *rs,
		    struct raid_set **ss, struct raid_dev *rd)
{
	struct lsi *lsi = META(rd, lsi);

	if (!init_set(lc, rs, rd, lsi->stride, type(lsi), handler))
		return 0;

	list_add_sorted(lc, &rs->devs, &rd->devs, dev_sort);

	switch (lsi->type) {
	case LSI_T_RAID0:
	case LSI_T_RAID1:
		break;

	case LSI_T_RAID10:
		/* FIXME: support RAID10. */
		LOG_ERR(lc, 0, "%s: RAID10 not supported yet", handler);

		if (!(*ss = join_superset(lc, name, super_created,
					  set_sort, rs, rd)))
			return 0;
	}

	return 1;
}

static struct raid_set *lsi_group(struct lib_context *lc,
				    struct raid_dev *rd)
{
	struct raid_set *rs, *ss = NULL;

	if ((rs = find_or_alloc_raid_set(lc, rd->name, FIND_TOP, rd,
					 LC_RS(lc), NO_CREATE, NO_CREATE_ARG)))
		return group_rd(lc, rs, &ss, rd) ? (ss ? ss : rs) : NULL;

	return NULL;
}

/*
 * Check an LSI RAID set.
 *
 * FIXME: add checks!!!
 */
static void _lsi_check(struct lib_context *lc, struct raid_set *rs)
{
	rs->status = s_ok;
}

static int lsi_check(struct lib_context *lc, struct raid_set *rs)
{
	check_set(lc, rs, lsi_check, handler);

	_lsi_check(lc, rs);

	return set_rs_status(lc, rs);
}

/*
 * IO error event handler.
 */
static int event_io(struct lib_context *lc, struct event_io *e_io)
{
	struct raid_dev *rd = e_io->rd;
	struct lsi *lsi = META(rd, lsi);

	/* Avoid write trashing. */
	if (status(lsi) & s_broken)
		return 0;

	// FIXME: lsi->what = MAGIC_BAD;

	return 1;
}

static struct event_handlers lsi_event_handlers = {
	.io = event_io,
	.rd = NULL,	/* FIXME: no device add/remove event handler yet. */
};

#ifdef DMRAID_NATIVE_LOG
/* Log native information about an LSI Logic RAID device. */
static void lsi_log(struct lib_context *lc, struct raid_dev *rd)
{
	unsigned int i;
	struct lsi *lsi = META(rd, lsi);

	log_print(lc, "%s (%s):", rd->di->path, handler);
	DP("magic_name: \"%s\"", lsi, lsi->magic_name);
	DP("seqno: %u", lsi, lsi->seqno);
	DP("type: %u", lsi, lsi->type);
	DP("stride: %u", lsi, lsi->stride);

	for (i = 0; i < 4; i++) {
		P("disks[%u].unknown: %u, 0x%x", lsi, lsi->disks[i].unknown, i,
		  lsi->disks[i].unknown, lsi->disks[i].unknown);
		P("disks[%u].magic_0: 0x%x, %x, %x", lsi,
		  lsi->disks[i].magic_0, i,
		  lsi->disks[i].magic_0,
		  (unsigned char) (((char*) &lsi->disks[i].magic_0)[0]),
		  (unsigned char) (((char*) &lsi->disks[i].magic_0)[1]));
		P("disks[%u].magic_1: 0x%x, %x, %x", lsi,
		  lsi->disks[i].magic_1, i,
		  lsi->disks[i].magic_1,
		  (unsigned char) (((char*) &lsi->disks[i].magic_1)[0]),
		  (unsigned char) (((char*) &lsi->disks[i].magic_1)[1]));
		P("disks[%u].disk_number: %u", lsi, lsi->disks[i].disk_number,
		  i, lsi->disks[i].disk_number);
		P("disks[%u].set_number: %u", lsi, lsi->disks[i].set_number,
		  i, lsi->disks[i].set_number);
	}
	DP("disk_number: %u", lsi, lsi->disk_number);
	DP("set_number: %u", lsi, lsi->set_number);
	P("what: %u, 0x%x", lsi, lsi->what, lsi->what, lsi->what);
}
#endif

static struct dmraid_format lsi_format = {
	.name	= HANDLER,
	.descr	= "(0) LSI Logic MegaRAID",
	.format = FMT_RAID,
	.read	= lsi_read,
	.write	= lsi_write,
	.group	= lsi_group,
	.check	= lsi_check,
	.events	= &lsi_event_handlers,
#ifdef DMRAID_NATIVE_LOG
	.log	= lsi_log,
#endif
};

/* Register this format handler with the format core */
int register_lsi(struct lib_context *lc)
{
	return register_format_handler(lc, &lsi_format);
}

/* Set the RAID device contents up derived from the LSI ones */
static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct lsi *lsi, struct dev_info *di)
{
	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas->offset = LSI_CONFIGOFFSET >> 9;
	rd->meta_areas->size   = sizeof(*lsi);
	rd->meta_areas->area   = (void*) lsi;

        rd->di = di;
	rd->fmt = &lsi_format;

	rd->offset = LSI_DATAOFFSET;
	rd->sectors = di->sectors - 1; /* FIXME: propper size ? */
	rd->sectors -= rd->sectors & 1;

	rd->status = status(lsi);
	rd->type   = type(lsi);

        if ((rd->name = name(lc, rd, 1)))
		return 1;

	return 0;
}
