/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Promise FastTrak ATARAID metadata format handler.
 *
 * pdc_read(), pdc_group() and group_rd() profited from
 * Carl-Daniel Hailfinger's raiddetect code.
 */

#define	HANDLER	"pdc"

#include "internal.h"
#define	FORMAT_HANDLER
#include "pdc.h"
#include <datastruct/byteorder.h>

static const char *handler = HANDLER;

/* Make up RAID device name. */
/* FIXME: better name ? */
static size_t _name(struct pdc *pdc, char *str, size_t len)
{
	return snprintf(str, len, "pdc_%u", pdc->magic_1) + 1;
}

static char *name(struct lib_context *lc, struct pdc *pdc, int type)
{
	size_t len;
	char *ret = NULL;

	if ((ret = dbg_malloc((len = _name(pdc, ret, 0)) + 1))) {
		_name(pdc, ret, len);
		mk_alpha(lc, ret + HANDLER_LEN, len - HANDLER_LEN);
	} else
		log_string_err(lc, handler);

	return ret;
}

/*
 * Retrieve status of device.
 *
 * FIXME: is this sufficient to cover all state ?
 */
static enum status status(struct pdc *pdc)
{
	return PDC_BROKEN(pdc) ? s_broken : s_ok;
}

/* Mapping of Promise types to generic types */
static struct types types[] = {
        { PDC_T_SPAN,   t_linear},
        { PDC_T_RAID0, t_raid0},
        { PDC_T_RAID1, t_raid1},
        { 0, t_undef}
};

/* Neutralize disk type */
static enum type type(struct pdc *pdc)
{
	return rd_type(types, (unsigned int) pdc->raid.type);
}

/* Calculate checksum on Promise metadata. */
static uint32_t checksum(struct pdc *pdc)
{
	unsigned int i = 511, sum = 0;
	uint32_t *p = (uint32_t*) pdc;

	while (i--)
		sum += *p++;

	return sum == pdc->checksum;
}

/* Calculate metadata offset. */
static loff_t offset(uint64_t size, unsigned int sectors)
{
	return (size - sectors) << 9;
}

/*
 * Read a Promise FastTrak RAID device
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct pdc *pdc)
{
	struct pdc_disk *disk;

	CVT32(pdc->magic_0);
	CVT32(pdc->magic_1);
	CVT32(pdc->raid.flags);
	CVT32(pdc->raid.magic_0);
	CVT32(pdc->raid.disk_secs);
	CVT32(pdc->raid.total_secs);
	CVT16(pdc->raid.cylinders);
	CVT32(pdc->raid.magic_1);

	for (disk = pdc->raid.disk; disk < &pdc->raid.disk[8]; disk++) {
		CVT32(disk->magic_0);
		CVT32(disk->disk_number);
	}
}
#endif

static int discover_signature(struct lib_context *lc, struct dev_info *di,
			      struct pdc *pdc, unsigned int *sector)
{
	static unsigned int *s, sectors[] = {
		PDC_CONFIGOFFSET,
		255,
		256,
		16,
		0
	};

	/* Assume certain track lengths. */
	for (s = sectors; *s; s++) {
		if (read_file(lc, handler, di->path, (void*) pdc, sizeof(*pdc),
			      offset(di->sectors, *s)) &&
        	    !strncmp(pdc->promise_id, PDC_MAGIC,
			     sizeof(PDC_MAGIC) - 1)) {
			*sector = *s;
			return 1;
		}
	}

	return 0;
}

/* Magic check. */
static int is_pdc(struct pdc *pdc)
{
	return pdc->magic_0 == pdc->raid.magic_0 &&
	       pdc->raid.total_disks < 8;
}

static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct pdc *pdc, struct dev_info *di,
		    unsigned int meta_sector);
static struct raid_dev *pdc_read(struct lib_context *lc, struct dev_info *di)
{ 
	unsigned int sector;
	struct pdc *pdc;
	struct raid_dev *rd = NULL;

	if (!(pdc = alloc_private(lc, handler, sizeof(*pdc))))
		goto out;

	if (!discover_signature(lc, di, pdc, &sector))
		goto bad;

	to_cpu(pdc);
	file_metadata(lc, handler, di->path, pdc, sizeof(*pdc),
		      offset(di->sectors, sector));

	if (!is_pdc(pdc)) {
		log_err(lc, "%s: identifying %s, magic_0: 0x%x/0x%x, "
			"magic_1: 0x%x/0x%x, total_disks: %u",
			handler, di->path,
			pdc->magic_0, pdc->raid.magic_0,
			pdc->magic_1, pdc->raid.magic_1, pdc->raid.total_disks);
		goto bad;
	}

	if (!checksum(pdc)) {
		log_err(lc, "%s: invalid checksum on %s", handler, di->path);
		goto bad;
	}

	if (!(rd = alloc_raid_dev(lc, handler)))
		goto bad;

	if (setup_rd(lc, rd, pdc, di, sector))
		goto out;

	log_err(lc, "%s: setting up RAID device %s", handler, di->path);
	free_raid_dev(lc, &rd);
	goto out;

   bad:
	dbg_free(pdc);
   out:
	return rd;
}

/*
 * Write a Promise FastTrak RAID device.
 */
static int pdc_write(struct lib_context *lc, struct raid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct pdc *pdc = META(rd, pdc);
#endif

	to_disk(pdc);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(pdc);
	return ret;
}

/* Decide about ordering sequence of RAID device. */
static int dev_sort(struct list_head *pos, struct list_head *new)
{
	return (META(RD(new), pdc))->raid.disk_number <
	       (META(RD(pos), pdc))->raid.disk_number;
}

/*
 * Group the RAID disk into a Promise set.
 */
static unsigned int stride(struct pdc *pdc)
{
	return pdc->raid.raid0_shift ? 1 << pdc->raid.raid0_shift : 0;
}

/* Add a PDC RAID device to a set. */
static struct raid_set *pdc_group(struct lib_context *lc,
				    struct raid_dev *rd)
{
	struct raid_set *rs;

	if ((rs = find_or_alloc_raid_set(lc, rd->name, FIND_ALL, rd, LC_RS(lc),
					 NO_CREATE, NO_CREATE_ARG))) {
		struct pdc *pdc = META(rd, pdc);

		if (init_set(lc, rs, rd, stride(pdc), pdc->raid.type, handler))
			list_add_sorted(lc, &rs->devs, &rd->devs, dev_sort);
		else
			rs = NULL;
	}

	return rs;
}

/*
 * Check a PDC RAID set.
 *
 * FIXME: more sanity checks.
 */
static void _pdc_check(struct lib_context *lc, struct raid_set *rs)
{
	unsigned int devs = count_devs(lc, rs, ct_dev);
	struct raid_dev *rd;

	list_for_each_entry(rd, &rs->devs, devs) {
		unsigned int disks;
		struct pdc *pdc = META(rd, pdc);

		log_dbg(lc, "checking %s device \"%s\"", handler, rd->di->path);
		switch (pdc->raid.type) {
		case PDC_T_SPAN:
		case PDC_T_RAID0:
		case PDC_T_RAID1:
			disks = pdc->raid.total_disks;
			break;

		default:
			disks = 0;
		}

		if (devs != disks) {
			rd->status = s_broken;
			log_err(lc, "%s: device %s broken in RAID "
				"set \"%s\" [%u/%u]",
				handler, rd->di->path, rs->name, devs, disks);
		}
	}
}

static int pdc_check(struct lib_context *lc, struct raid_set *rs)
{
	_pdc_check(lc, rs);
	return set_rs_status(lc, rs);
}

/*
 * IO error event handler.
 */
static int event_io(struct lib_context *lc, struct event_io *e_io)
{
	struct raid_dev *rd = e_io->rd;
	struct pdc *pdc = META(rd, pdc);

	/* Avoid write trashing. */
	if (status(pdc) & s_broken)
		return 0;

	PDC_SET_BROKEN(pdc);

	return 1;
}

static struct event_handlers pdc_event_handlers = {
	.io = event_io,
	.rd = NULL,	/* FIXME: no device add/remove event handler yet. */
};

#ifdef DMRAID_NATIVE_LOG
/* Log native information about a Promise RAID device. */
static void pdc_log(struct lib_context *lc, struct raid_dev *rd)
{
	unsigned int i;
	struct pdc *pdc = META(rd, pdc);

	log_print(lc, "%s (%s):", rd->di->path, handler);
	DP("promise_id: \"%s\"", pdc, pdc->promise_id);
	DP("unknown_0: 0x%x", pdc, pdc->unknown_0);
	DP("magic_0: 0x%x", pdc, pdc->magic_0);
	DP("unknown_1: 0x%x", pdc, pdc->unknown_1);
	DP("magic_1: 0x%x", pdc, pdc->magic_1);
	DP("unknown_2: 0x%x", pdc, pdc->unknown_2);

	DP("raid.flags: 0x%x", pdc, pdc->raid.flags);
	DP("raid.unknown_0: 0x%x", pdc, pdc->raid.unknown_0);
	DP("raid.disk_number: %u", pdc, pdc->raid.disk_number);
	DP("raid.channel: %u", pdc, pdc->raid.channel);
	DP("raid.device: %u", pdc, pdc->raid.device);
	DP("raid.magic_0: 0x%x", pdc, pdc->raid.magic_0);
	DP("raid.unknown_1: 0x%x", pdc, pdc->raid.unknown_1);
	DP("raid.unknown_2: 0x%x", pdc, pdc->raid.unknown_2);
	DP("raid.disk_secs: %u", pdc, pdc->raid.disk_secs);
	DP("raid.unknown_3: 0x%x", pdc, pdc->raid.unknown_3);
	DP("raid.unknown_4: 0x%x", pdc, pdc->raid.unknown_4);
	DP("raid.status: 0x%x", pdc, pdc->raid.status);
	DP("raid.type: 0x%x", pdc, pdc->raid.type);
	DP("raid.total_disks: %u", pdc, pdc->raid.total_disks);
	DP("raid.raid0_shift: %u", pdc, pdc->raid.raid0_shift);
	DP("raid.raid0_disks: %u", pdc, pdc->raid.raid0_disks);
	DP("raid.array_number: %u", pdc, pdc->raid.array_number);
	DP("raid.total_secs: %u", pdc, pdc->raid.total_secs);
	DP("raid.cylinders: %u", pdc, pdc->raid.cylinders);
	DP("raid.heads: %u", pdc, pdc->raid.heads);
	DP("raid.sectors: %u", pdc, pdc->raid.sectors);
	DP("raid.magic_1: 0x%x", pdc, pdc->raid.magic_1);
	DP("raid.unknown_5: 0x%x", pdc, pdc->raid.unknown_5);

	for (i = 0; i < 8; i++) {
		if (!pdc->raid.disk[i].magic_0)
			continue;

		P2("raid.disk[%d].unknown_0: 0x%x", pdc,
		   i, pdc->raid.disk[i].unknown_0);
		P2("raid.disk[%d].channel: %u", pdc,
		   i, pdc->raid.disk[i].channel);
		P2("raid.disk[%d].device: %u", pdc,
		   i, pdc->raid.disk[i].device);
		P2("raid.disk[%d].magic_0: 0x%x", pdc,
		   i, pdc->raid.disk[i].magic_0);
		P2("raid.disk[%d].disk_number: %u", pdc,
		   i, pdc->raid.disk[i].disk_number);
	}

	P("checksum: 0x%x %s", pdc, pdc->checksum, pdc->checksum,
	   checksum(pdc) ? "Ok" : "Crap");
}
#endif

static struct dmraid_format pdc_format = {
	.name	= HANDLER,
	.descr	= "(+) Promise FastTrack",
	.format = FMT_RAID,
	.read	= pdc_read,
	.write	= pdc_write,
	.group	= pdc_group,
	.check	= pdc_check,
	.events	= &pdc_event_handlers,
#ifdef DMRAID_NATIVE_LOG
	.log	= pdc_log,
#endif
};

/* Register this format handler with the format core */
int register_pdc(struct lib_context *lc)
{
	return register_format_handler(lc, &pdc_format);
}

static uint64_t sectors(struct raid_dev *rd, unsigned int meta_sector)
{
	uint64_t ret;
	struct pdc *pdc = META(rd, pdc);

	switch (pdc->raid.type) {
	case PDC_T_SPAN:
		return rd->di->sectors - meta_sector;

	case PDC_T_RAID0:
	case PDC_T_RAID1:
		ret = pdc->raid.total_secs / pdc->raid.total_disks;
		if (ret < rd->di->sectors - meta_sector)
			return ret;
	}

	return 0;
}

/* Set the RAID device contents up derived from the PDC ones */
static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct pdc *pdc, struct dev_info *di,
		    unsigned int meta_sector)
{
	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas->offset = di->sectors - meta_sector;
	rd->meta_areas->size = sizeof(*pdc);
	rd->meta_areas->area = (void*) pdc;

        rd->di = di;
	rd->fmt = &pdc_format;

	rd->offset = PDC_DATAOFFSET;
	if (!(rd->sectors = sectors(rd, meta_sector)))
		return 0;

	rd->status = status(pdc);
	rd->type   = type(pdc);

        if ((rd->name = name(lc, pdc, 0)))
		return 1;

	return 0;
}
