/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Silicon Image Medley ATARAID metadata format handler.
 */

#define	HANDLER	"sil"

#include "internal.h"
#define	FORMAT_HANDLER
#include "sil.h"
#include <datastruct/byteorder.h>

static const char *handler = HANDLER;

/* Make up RAID device name from some 'magic' numbers */
/* FIXME: better name ? */
static size_t _name(struct sil *sil, char *str, size_t len,
		     unsigned int subset)
{
	const char *fmt = subset ? "sil_%02u%02u%02u%02u%02u%02u%u-%u" :
				   "sil_%02u%02u%02u%02u%02u%02u%u";

	return snprintf(str, len, fmt,
			sil->year, sil->month, sil->day,
			sil->hour, sil->minutes % 60, sil->seconds % 60,
			sil->incarnation_no, sil->mirrored_set_number) + 1;
}

static char *name(struct lib_context *lc, struct raid_dev *rd,
		  unsigned int subset)
{
	size_t len;
	char *ret;
	struct sil *sil = META(rd, sil);

	subset = subset && sil->type == SIL_T_RAID10;
	if ((ret = dbg_malloc((len = _name(sil, NULL, 0, subset)) + 1))) {
		_name(sil, ret, len, subset);
		mk_alpha(lc, ret + HANDLER_LEN, len - HANDLER_LEN -
			 (strrchr(ret, '-') ? 3 : 1));
	} else
		log_string_err(lc, handler);

	return ret;
}

/*
 * Retrieve status of device.
 * FIXME: is this sufficient to cover all state ?
 */
static enum status status(struct sil *sil)
{
	return sil->member_status == 1 ? s_ok : s_broken;
}

/* Mapping of SIL 680 types to generic types */
static struct types types[] = {
        { SIL_T_RAID0,  t_raid0},
        { SIL_T_RAID1,  t_raid1},
        { SIL_T_RAID10, t_raid0},
        { SIL_T_SPARE,   t_spare},
        { 0, t_undef}
};

/* Neutralize disk type */
static enum type type(struct sil *sil)
{
	return rd_type(types, (unsigned int) sil->type);
}

/* Calculate checksum on metadata */
static uint16_t _checksum(struct sil *sil, unsigned int count)
{
	uint16_t *p = (uint16_t*) sil, sum = 0;

        while (count--)
		sum += *p++;

	return sum;
}

/* Calculate all checksums on metadata */
/* FIXME: checksum1 check correct ? */
static uint32_t checksum(struct sil* sil)
{
	/* FIXME: do we need this check ?
	    _checksum((struct sil*) &sil->auto_rebuild,
		      (unsigned int) ((unsigned long) &sil->checksum1 -
		      (unsigned long) &sil->auto_rebuild)) == sil->checksum1
	*/

	if (!_checksum(sil, 160) &&
	    _checksum(sil, 256) == sil->checksum2)
		return 1;


	return 0;
}

/*
 * Read a Silicon Image RAID device
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct sil *sil)
{
	CVT32(sil->magic);
	CVT32(sil->array_sectors_low);
	CVT32(sil->array_sectors_high);
	CVT32(sil->thisdisk_sectors);
	CVT16(sil->product_id);
	CVT16(sil->vendor_id);
	CVT16(sil->minor_ver);
	CVT16(sil->major_ver);
	CVT16(sil->raid0_stride);
	CVT32(sil->rebuild_ptr_low);
	CVT32(sil->rebuild_ptr_high);
	CVT32(sil->incarnation_no);
	CVT16(sil->checksum1);
	CVT16(sil->checksum2);
}
#endif

static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct sil *sil, struct dev_info *di);
#define	AREAS	4
#define SIL_META_AREA(i)	(SIL_CONFIGOFFSET - (i * 512 << 9))

static int is_sil(struct sil *sil)
{
	return sil->magic == SIL_MAGIC && sil->disk_number < 8;
}

static int sil_valid(struct lib_context *lc, struct dev_info *di,
		     struct sil *sil)
{
	if (!is_sil(sil))
		return 0;

	if (sil->major_ver != 2)
		log_warn(lc, "%s: major version %u; format handler "
			 "tested on version 2 only", handler, sil->major_ver);

	if (!checksum(sil))
		LOG_ERR(lc, 0, "%s: invalid metadata checksum on %s",
			handler, di->path);

	if (di->sectors < sil->thisdisk_sectors)
		LOG_ERR(lc, 0, "%s: invalid disk size in metadata on %s",
			handler, di->path);

	return 1;
}

static int _sil_read(struct lib_context *lc, struct dev_info *di,
		     struct sil *sils[])
{
	unsigned int i, valid;
	char str[9] = { 0, };
	struct sil *sil;

	/* Read the 4 metadata areas. */
	for (i = valid = 0; i < AREAS; i++) {
		if (!(sil = alloc_private_and_read(lc, handler, sizeof(*sil),
						   di->path, SIL_META_AREA(i))))
			return 0;

		to_cpu(sil);
		sils[i] = sil;

		/* Create string with list of valid areas. */
		if (sil_valid(lc, di, sil))
			sprintf(&str[strlen(str)], "%s%u",
			        valid++ ? "," : "", i + 1);
	}

	if (valid)
		log_info(lc, "%s: area%s %s[%u] %s valid",
			 handler, valid ? "s" : "", str, AREAS,
			 valid == 1 ? "is" : "are");
	
	return valid ? 1 : 0;
}

static int _file_name(char *str, size_t len, char *n, int i)
{
	return snprintf(str, len, "%s_%d", n, i) + 1;
}

static char *file_name(struct lib_context *lc, char *n, int i)
{
	size_t len;
	char *ret;

	if ((ret = dbg_malloc((len = _file_name(NULL, 0, n, i)))))
		_file_name(ret, len, n, i);
	else
		log_string_err(lc, handler);

	return ret;
}

static void _file_metadata(struct lib_context *lc, struct dev_info *di,
			   struct sil *sils[])
{
	unsigned int i;
	char *n;

	for (i = 0; i < AREAS; i++) {
		if (!(n = file_name(lc, di->path, i)))
			break;

		file_metadata(lc, handler, n, sils[i],
			      sizeof(**sils), SIL_META_AREA(i) >> 9);

		dbg_free(n);
	}
}

/* Quorate SIL metadata copies. */
static struct sil *_elect(struct lib_context *lc, struct dev_info *di,
			  struct sil *sils[])
{
	unsigned int areas, i, j, q = 0;
	unsigned char m, old_m = 0;
	struct sil *sil, *sil1;
	
	for (areas = i = 0; i < AREAS; i++) {
		if (sil_valid(lc, di, sils[i]))
			areas++;
		else {
			dbg_free(sils[i]);
			sils[i] = NULL;
		}
	}

	if (areas == AREAS)
		goto out;

	log_err(lc, "%s: only %u/%u metadata areas found on %s, %sing...",
		handler, areas, AREAS, di->path,
		areas > 1 ? "elect" : "pick");
	for (i = 0; i < AREAS; i++) {
		if ((sil = sils[i])) {
			for (m = 0, j = i + 1; j < AREAS; j++) {
				if ((sil1 = sils[j]) &&
				    !memcmp(sil, sil1, sizeof(*sil)))
					m++;
			}

			if (m >= old_m) {
				q = i;
				old_m = m;
			}
		}
	}

	if (q) {
		dbg_free(sils[0]);
		sils[0] = sils[q];
		sils[q] = NULL;
	}

   out:
	return sils[0];
}

static struct raid_dev *sil_read(struct lib_context *lc, struct dev_info *di)
{
	unsigned int i = 0;
	struct sil *sils[AREAS];
	struct raid_dev *rd = NULL;

	/* Read all 4 metadata areas off disk. */
	memset(sils, 0, sizeof(sils));
	if (!_sil_read(lc, di, sils))
		goto bad;

	_file_metadata(lc, di, sils);

	if (!(rd = alloc_raid_dev(lc, handler)))
		goto bad;

	if (setup_rd(lc, rd, _elect(lc, di, sils), di)) {
		i = 1;
		goto bad;
	}

	log_err(lc, "%s: setting up RAID device %s", handler, di->path);
	free_raid_dev(lc, &rd);
	goto out;

   bad:
	for (; i < AREAS; i++) {
		if (sils[i])
			dbg_free(sils[i]);
	}
   out:
	return rd;
}

/*
 * Write a Silicon Image RAID device.
 */
static int sil_write(struct lib_context *lc, struct raid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct sil *sil = META(rd, sil);
#endif

	to_disk(sil);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(sil);
	return ret;
}

/* Decide about ordering sequence of RAID device. */
static int dev_sort(struct list_head *pos, struct list_head *new)
{
	return (META(RD(new), sil))->disk_number <
	       (META(RD(pos), sil))->disk_number;
}

/* Decide about ordering sequence of RAID device. */
static int set_sort(struct list_head *pos, struct list_head *new)
{
	return (META(RD_RS(RS(new)), sil))->mirrored_set_number <
	       (META(RD_RS(RS(pos)), sil))->mirrored_set_number;
}

/*
 * Group the RAID disk into a SIL set.
 *
 * Check device hierarchy and create super set appropriately.
 */
static void super_created(struct raid_set *super, void *private)
{
	super->type   = t_raid1;
	super->stride = META((struct raid_dev*) private, sil)->raid0_stride;
}

/* FIXME: handle spares. */
static int group_rd(struct lib_context *lc, struct raid_set *rs,
		    struct raid_set **super_set, struct raid_dev *rd)
{
	struct sil *sil = META(rd, sil);

	if (!init_set(lc, rs, rd, sil->raid0_stride, sil->type, handler))
		return 0;

	list_add_sorted(lc, &rs->devs, &rd->devs, dev_sort);

	switch (sil->type) {
	case SIL_T_RAID0:
	case SIL_T_RAID1:
		if (!(find_set(lc, NULL, rs->name, FIND_TOP)))
			list_add_tail(&rs->list, LC_RS(lc));

		break;

	case SIL_T_RAID10:
		/*
		 * We've got a striped raid set with a mirror set on top
		 * when we get here.
		 * Let's find and optionally allocate the mirror set.
		 */
		if (!(*super_set = join_superset(lc, name, super_created,
						 set_sort, rs, rd)))
			return 0;

		break;

	case SIL_T_SPARE:
		break;

	default:
		return 0;
	}

	return 1;
}

/* Add a SIL RAID device to a set */
static struct raid_set *sil_group(struct lib_context *lc,
				    struct raid_dev *rd)
{
	struct raid_set *rs, *super = NULL;

	if ((rs = find_or_alloc_raid_set(lc, rd->name, FIND_ALL, rd,
					 NO_LIST, NO_CREATE, NO_CREATE_ARG)))
		return group_rd(lc, rs, &super, rd) ?
		       (super ? super : rs) : NULL;

	return NULL;
}

/*
 * Check a SIL RAID set.
 *
 * FIXME: more sanity checks.
 */
static void _sil_check(struct lib_context *lc, struct raid_set *rs)
{
	unsigned int devs;
	struct raid_dev *rd;

	if (!DEVS(rs))
		return;

	devs = count_devs(lc, rs, ct_dev);
	list_for_each_entry(rd, &rs->devs, devs) {
		unsigned char disks;
		struct sil *sil = META(rd, sil);

		log_dbg(lc, "checking %s device \"%s\"", handler, rd->di->path);
		switch (sil->type) {
		case SIL_T_RAID0:
		case SIL_T_RAID10:
			disks = sil->drives_per_striped_set;
			break;

		case SIL_T_RAID1:
			disks = sil->drives_per_mirrored_set;
			break;

		default:
			disks = 0;
		}

		if (devs != disks) {
			rd->status = s_inconsistent;
			log_err(lc, "%s: wrong # of devices in RAID "
				"set \"%s\" [%u/%u]",
				handler, rs->name, devs, disks);
		}
	}
}

static int sil_check(struct lib_context *lc, struct raid_set *rs)
{
	/* Check hierachical RAID sets */
	check_set(lc, rs, sil_check, handler);

	_sil_check(lc, rs);

	return set_rs_status(lc, rs);
}

/*
 * IO error event handler.
 */
static int event_io(struct lib_context *lc, struct event_io *e_io)
{
	struct raid_dev *rd = e_io->rd;
	struct sil *sil = META(rd, sil);

	/* Avoid write trashing. */
	if (status(sil) & s_broken)
		return 0;

	sil->member_status = 0;

	return 1;
}

static struct event_handlers sil_event_handlers = {
	.io = event_io,
	.rd = NULL,	/* FIXME: no device add/remove event handler yet. */
};

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about a Silicon Image  RAID device.
 */
static void sil_log(struct lib_context *lc, struct raid_dev *rd)
{
	struct sil *sil = META(rd, sil);

	log_print(lc, "%s (%s):", rd->di->path, handler);
	DP("unknown0: \"%42s\"", sil, sil->unknown0);
	DP("ascii_version: \"%8s\"", sil, sil->ascii_version);
	DP("diskname: \"%32s\"", sil, sil->diskname);
	DP("unknown1: \"%22s\"", sil, sil->unknown1);
	DP("magic: 0x%x", sil, sil->magic);
	DP("unknown1a: \"%8s\"", sil, sil->unknown1a);
	DP("array_sectors_low: %u", sil, sil->array_sectors_low);
	DP("array_sectors_high: %u", sil, sil->array_sectors_high);
	DP("unknown2: \"%4s\"", sil, sil->unknown2);
	DP("thisdisk_sectors: %u", sil, sil->thisdisk_sectors);
	DP("product_id: %u", sil, sil->product_id);
	DP("vendor_id: %u", sil, sil->vendor_id);
	DP("minor_ver: %u", sil, sil->minor_ver);
	DP("major_ver: %u", sil, sil->major_ver);
	DP("seconds: %u", sil, sil->seconds % 60);
	DP("seconds(full): 0x%x", sil, sil->seconds);
	DP("minutes: %u", sil, sil->minutes % 60);
	DP("minutes(full): 0x%x", sil, sil->minutes);
	DP("hour: %u", sil, sil->hour);
	DP("day: %u", sil, sil->day);
	DP("month: %u", sil, sil->month);
	DP("year: %u", sil, sil->year);
	DP("raid0_stride: %u", sil, sil->raid0_stride);
	DP("disk_number: %u", sil, sil->disk_number);
	DP("type: %u", sil, sil->type);
	DP("drives_per_striped_set: %d", sil, sil->drives_per_striped_set);
	DP("striped_set_number: %d", sil, sil->striped_set_number);
	DP("drives_per_mirrored_set: %d", sil, sil->drives_per_mirrored_set);
	DP("mirrored_set_number: %d", sil, sil->mirrored_set_number);
	DP("rebuild_ptr_low: %u", sil, sil->rebuild_ptr_low);
	DP("rebuild_ptr_high: %u", sil, sil->rebuild_ptr_high);
	DP("incarnation_no: %u", sil, sil->incarnation_no);
	DP("member_status: %u", sil, sil->member_status);
	DP("mirrored_set_state: %u", sil, sil->mirrored_set_state);
	DP("reported_device_location: %u", sil, sil->reported_device_location);
	DP("idechannel: %u", sil, sil->idechannel);
	DP("auto_rebuild: %u", sil, sil->auto_rebuild);
	DP("checksum1: %u", sil, sil->checksum1);
	DP("checksum2: %u", sil, sil->checksum2);
}
#endif

static struct dmraid_format sil_format = {
	.name	= HANDLER,
	.descr	= "(+) Silicon Image(tm) Medley(tm)",
	.format = FMT_RAID,
	.read	= sil_read,
	.write	= sil_write,
	.group	= sil_group,
	.check	= sil_check,
	.events	= &sil_event_handlers,
#ifdef DMRAID_NATIVE_LOG
	.log	= sil_log,
#endif
};

/* Register this format handler with the format core */
int register_sil(struct lib_context *lc)
{
	return register_format_handler(lc, &sil_format);
}

/* Set the RAID device contents up derived from the SIL ones */
static int stripes(struct sil *sil)
{
	return sil->drives_per_striped_set > -1 &&
	       sil->disk_number < sil->drives_per_striped_set;
}

static int mirrors(struct sil *sil)
{
	return sil->drives_per_mirrored_set > -1 &&
	       sil->disk_number < sil->drives_per_mirrored_set;
}

static uint64_t sectors(struct raid_dev *rd)
{
	uint64_t array_sectors;
	struct sil *sil = META(rd, sil);

	array_sectors = (((uint64_t) sil->array_sectors_high) << 32) +
			 sil->array_sectors_low;

	switch (sil->type) {
	case SIL_T_RAID0:
		if (stripes(sil))
			return array_sectors / sil->drives_per_striped_set;

		break;

	case SIL_T_RAID1:
		if (mirrors(sil))
			return array_sectors;

		break;

	case SIL_T_RAID10:
		if (mirrors(sil) && stripes(sil))
			return array_sectors / sil->drives_per_mirrored_set;

		break;

	case SIL_T_SPARE:
		/* Cook them up... */
		return rd->di->sectors - (AREAS - 1) * 512 -
		       ((rd->di->sectors & 1) ? 1 : 2);
	}

	return 0;
}

static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct sil *sil, struct dev_info *di)
{
	unsigned int i;
	struct meta_areas *ma;

	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, AREAS)))
		return 0;

	for (i = 0, ma = rd->meta_areas; i < rd->areas; i++, ma++) {
		ma->offset = SIL_META_AREA(i) >> 9;
		ma->size = sizeof(*sil);
		ma->area = (void*) sil;
	}

	rd->di = di;
	rd->fmt = &sil_format;

	rd->offset = SIL_DATAOFFSET;
	if (!(rd->sectors = sectors(rd)))
		return 0;

	rd->status = status(sil);
	rd->type   = type(sil);

        if ((rd->name = name(lc, rd, sil->type == SIL_T_RAID10)))
		return 1;

	return 0;
}
