/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include "internal.h"
#include "ondisk.h"

/*
 * Metadata format handler registry.
 */

/*
 * Used for development.
 *
 * Comment next line out to avoid pre-registration
 * checks on metadata format handlers.
 */
#define	CHECK_FORMAT_HANDLER
#ifdef	CHECK_FORMAT_HANDLER
/*
 * Check that mandatory members of a metadata form handler are present.
 * 
 * We can only use log_err and log_print here, because debug and verbose
 * options are checked for later during initialization...
 */

/*
 * Because we have a bunch of members to check,
 * let's define them as an array.
 */
#define offset(member)	((unsigned short) &((struct dmraid_format *)0)->member)

struct format_member {
	const unsigned short offset;
	const unsigned char all;
	const unsigned char method;
	const char *msg;
} __attribute((packed));

static struct format_member format_members[] = {
	{ offset(name),   1, 0, "name" },
	{ offset(descr),  1, 0, "description" },
	{ offset(read),   1, 1, "read" },
	{ offset(write),  0, 1, "write" },
	{ offset(group),  1, 1, "group" },
	{ offset(check),  1, 1, "check" },
	{ offset(events), 0, 0, "events array" },
#ifdef	NATIVE_LOG
	{ offset(log),    0, 1, "log" },
#endif
};
#undef	offset

static int check_member(struct lib_context *lc, struct dmraid_format *fmt,
			struct format_member *member)
{
	if ((!member->all && fmt->format != FMT_RAID) ||
	    *((unsigned long*) (((unsigned char*) fmt) + member->offset)))
		return 0;

	LOG_ERR(lc, 1, "%s: missing metadata format handler %s%s",
		fmt->name, member->msg, member->method ? " method" : "");
}

static int check_format_handler(struct lib_context *lc,
				struct dmraid_format *fmt)
{
	unsigned int error, i;

	if (!fmt)
		BUG(lc, 0, "NULL metadata format handler");

	for (error = i = 0; i < ARRAY_SIZE(format_members); i++)
		error += check_member(lc, fmt, &format_members[i]);

	return error ? 0 : 1;
}
#endif /* CHECK_FORMAT_HANDLER */

/*
 * Register a RAID metadata format handler.
 */
int register_format_handler(struct lib_context *lc, struct dmraid_format *fmt)
{
	struct format_list *fl;

#ifdef	CHECK_FORMAT_HANDLER
	if (!check_format_handler(lc, fmt))
		return 0;
#undef	CHECK_FORMAT_HANDLER
#endif
	if (!(fl = dbg_malloc(sizeof(*fl))))
		return 0;

	fl->fmt = fmt;
	list_add_tail(&fl->list, LC_FMT(lc));

	return 1;
}

/*
 * (Un)Register all format handlers.
 *
 * I use an array because of the growing number...
 */
static struct register_fh {
	int (*func)(struct lib_context *lc);
} register_fh[] = {
#include "register.h"
	{ NULL },
};

void unregister_format_handlers(struct lib_context *lc)
{
	struct list_head *elem, *tmp;

	list_for_each_safe(elem, tmp, LC_FMT(lc)) {
		list_del(elem);
		dbg_free(list_entry(elem, typeof(struct format_list), list));
	}
}

int register_format_handlers(struct lib_context *lc)
{
	int ret = 1;
	struct register_fh *fh;

	for (fh = register_fh; fh->func; fh++) {
		if ((ret = fh->func(lc)))
			continue;

		/* Clean up in case of error */
		log_err(lc, "registering format");
		unregister_format_handlers(lc);
		break;
	}

	return ret;
}
/* END metadata format handler registry. */


/*
 * Other metadata format handler support functions.
 */

/* Allocate private space in format handlers (eg, for on-disk metadata). */
void *alloc_private(struct lib_context *lc, const char *who, size_t size)
{
	void *ret;

	if (!(ret = dbg_malloc(size)))
		log_err(lc, "allocating %s metadata", who);

	return ret;
}

/* Allocate private space in format handlers and read data off device. */
void *alloc_private_and_read(struct lib_context *lc, const char *who,
			     size_t size, char *path, loff_t offset)
{
	void *ret;

	if ((ret = alloc_private(lc, who, size))) {
		if (!read_file(lc, who, path, ret, size, offset)) {
			dbg_free(ret);
			ret = NULL;
		}
	}

	return ret;
}


/* Allocate metadata sector array in format handlers. */
void *alloc_meta_areas(struct lib_context *lc, struct raid_dev *rd,
		       const char *who, unsigned int n)
{
	void *ret;

	if ((ret = alloc_private(lc, who, n * sizeof(*rd->meta_areas))))
		rd->areas = n;

	return ret;
}

/* Simple metadata write function for format handlers */
static int _write_metadata(struct lib_context *lc, const char *handler,
			   struct raid_dev *rd, int idx, int erase)
{
	int ret = 0;
	void *p, *tmp;

	if (idx >= rd->areas)
		goto out;

	p = tmp = rd->meta_areas[idx].area;
	if (erase &&
	    !(p = alloc_private(lc, handler, rd->meta_areas[idx].size)))
		goto out;

	ret = write_file(lc, handler, rd->di->path, (void*) p,
			 rd->meta_areas[idx].size,
			 rd->meta_areas[idx].offset << 9);

	log_dbg(lc, "writing metadata to %s, offset %" PRIu64 " sectors, "
		"size %zu sectors returned %d",
		rd->di->path, rd->meta_areas[idx].offset,
		rd->meta_areas[idx].size, ret); 
	if (p != tmp)
		dbg_free(p);

  out:
	return ret;
}

int write_metadata(struct lib_context *lc, const char *handler,
		   struct raid_dev *rd, int idx, int erase)
{
	unsigned int i;

	if (idx > -1)
		return _write_metadata(lc, handler, rd, idx, erase);

	for (i = 0; i < rd->areas; i++) {
		if (!_write_metadata(lc, handler, rd, i, erase))
			return 0;
	}

	return 1;
}

/*
 * Check list of RAID sets.
 */
void check_set(struct lib_context *lc, struct raid_set *rs,
	       int (*check)(struct lib_context *lc, struct raid_set *r),
	       const char *handler)
{
	struct raid_set *r;
	struct raid_dev *rd;
	uint64_t sectors = total_sectors(lc, rs);

	list_for_each_entry(rd, &rs->devs, devs) {
		if (T_SPARE(rd) &&
		    rs->type == t_raid1 &&	/* FIXME: rs->type check ? */
		    rd->sectors != sectors)
			log_err(lc, "%s: size mismatch in set \"%s\", spare "
				"\"%s\"", handler, rs->name, rd->di->path);
	}

	list_for_each_entry(r, &rs->sets, list) {
		log_dbg(lc, "checking RAID set \"%s\"", r->name);
		check(lc, r);
	}
}

/* Initialize a RAID sets type and stride. */
int init_set(struct lib_context *lc, struct raid_set *rs,
	     struct raid_dev *rd, unsigned int stride,
	     unsigned int type, const char *handler)
{
	if (T_UNDEF(rd))
		LOG_ERR(lc, 0, "%s: RAID type %u not supported", handler, type);

	if (T_SPARE(rs) || T_UNDEF(rs))
		rs->type = rd->type;
	else if (!T_SPARE(rd) && rs->type != rd->type)
		log_err(lc, "%s: RAID type mismatch in \"%s\" on  %s",
			handler, rs->name, rd->di->path);

	if (rs->stride) {
		if (rs->stride != stride)
			LOG_ERR(lc, 0,
				"%s: stride inconsistency detected on \"%s\"",
				handler, rd->di->path);
	} else
		rs->stride = stride;

	return 1;
}

/* Check if format identifier is valid. */
int check_valid_format(struct lib_context *lc, char *name)
{
	struct format_list *fl;

	list_for_each_entry(fl, LC_FMT(lc), list) {
		if (!strncmp(name, fl->fmt->name, strlen(name)))
			return 1;
	}

	return 0;
}


/*
 * Update RAID set state based on operational subsets/devices.
 *
 * In case of a RAID set hierachy, check availability of subsets
 * and set superset to broken in case *all* subsets are broken.
 * If at least one is still available, set to inconsistent.
 *
 * In case of lowest level RAID sets, check consistence of devices
 * and make the above decision at the device level.
 */
static void _set_rs_status(struct lib_context *lc, struct raid_set *rs,
			   unsigned int i, unsigned int operational)
{
	if (operational == i)
		rs->status = s_ok;
	else if (operational)
		rs->status = s_inconsistent;
	else
		rs->status = s_broken;

	log_dbg(lc, "set status of set \"%s\" to %u", rs->name, rs->status);
}

int set_rs_status(struct lib_context *lc, struct raid_set *rs)
{
	unsigned int i = 0, operational = 0;
	struct raid_set *r;
	struct raid_dev *rd;

	/* Set status of subsets. */
	list_for_each_entry(r, &rs->sets, list) {
		/* Check subsets to set status of superset. */
		i++;
		if (S_OK(r->status) || S_INCONSISTENT(r->status))
			operational++;
	}
	
	/* Check status of devices... */
	list_for_each_entry(rd, &rs->devs, devs) {
		i++;
		if (S_OK(rd->status))
			operational++;
	}
	
	_set_rs_status(lc, rs, i, operational);

	return S_BROKEN(rs->status) ? 0 : 1;
}

/*
 * Allocate a RAID superset and link the subset to it.
 */
struct raid_set *join_superset(struct lib_context *lc,
			       char *(*f_name)(struct lib_context *lc,
					       struct raid_dev *rd,
					       unsigned int subset),
			       void (*f_create)(struct raid_set *super,
					        void *private),
			       int (*f_set_sort)(struct list_head *pos,
						 struct list_head *new),
			       struct raid_set *rs,
			       struct raid_dev *rd)
{
	char *n;
	struct raid_set *ret;

	if (!(n = f_name(lc, rd, 0)))
		return NULL;

	ret = find_or_alloc_raid_set(lc, n, FIND_TOP, NO_RD,
				     LC_RS(lc), f_create, rd);
	dbg_free(n);
	if (ret) {
		if (!find_set(lc, &ret->sets, rs->name, FIND_TOP))
			list_add_sorted(lc, &ret->sets, &rs->list, f_set_sort);
	}

	return ret;
}

