/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#ifndef	_FORMAT_H_
#define	_FORMAT_H_

/* Metadata format handler types. */
enum fmt_type {
	FMT_RAID,
	FMT_PARTITION,
};

#ifdef FORMAT_HANDLER
#undef FORMAT_HANDLER

#include <sys/types.h>
#include "datastruct/list.h"
#include "metadata/metadata.h"

/*
 * Various structures for data whipped with the different event types.
 */
/* Event: I/O error */
struct event_io {
	struct raid_set *rs;	/* RAID set of I/O error. */
	struct raid_dev *rd;	/* RAID device of I/O error. */
	uint64_t sector;	/* Sector of the I/O error. */
};

/* Event: RAID device add/remove */
enum rd_type {
	rd_add,
	rd_remove,
};

struct event_rd {
	struct raid_set *rs;
	struct raid_dev *rd;
	enum rd_type	type;
};

/*
 * List of event handler functions to call for the metadata format handler.
 *
 * Return 1 for event taken, RAID device write necessary.
 * Return 0 for error and/or write unnecessary.
 */
struct event_handlers {
	/* Handle IO error */
	int (*io)(struct lib_context *lc, struct event_io *e_io);

	/* Handle RAID device add/remove. */
	int (*rd)(struct lib_context *lc, struct event_rd *e_rd);
};

/*
 * Virtual interface definition of a metadata format handler.
 */
struct dmraid_format {
	const char *name;	/* Format name */
	const char *descr;	/* Format description */
	enum fmt_type format;	/* Format type (RAID, partition) */

	/*
	 * Read RAID metadata off a device and unify it.
	 */
	struct raid_dev* (*read)(struct lib_context *lc, struct dev_info* di);

	/*
	 * Write RAID metadata to a device deunifying it
	 * or erase ondisk metadata if erase != 0.
	 */
	int (*write)(struct lib_context *lc, struct raid_dev* rd, int erase);

	/*
	 * Group a RAID device into a set.
	 */
	struct raid_set* (*group)(struct lib_context *lc, struct raid_dev *rd);

	/*
	 * Check consistency of the RAID set metadata.
	 */
	int (*check)(struct lib_context *lc, struct raid_set *rs);

	/*
	 * Event handlers (eg, I/O error).
	 */
	struct event_handlers *events;

	/*
	 * Display RAID disk metadata native.
	 */
	void (*log)(struct lib_context *lc, struct raid_dev *rd);
};

/* Chain of registered format handlers (needed for library context). */
struct format_list {
	struct list_head list;
	struct dmraid_format *fmt;
};


void check_set(struct lib_context *lc, struct raid_set *rs,
	       int (*check)(struct lib_context *lc, struct raid_set *r),
	       const char *handler);
int check_valid_format(struct lib_context *lc, char *fmt);
int init_set(struct lib_context *lc, struct raid_set *rs,
	     struct raid_dev *rd, unsigned int stride,
	     unsigned int type, const char *handler);
int register_format_handlers(struct lib_context *lc);
void unregister_format_handlers(struct lib_context *lc);


/* All metadata format handlers use these. */
void *alloc_meta_areas(struct lib_context *lc, struct raid_dev *rd,
		       const char *who, unsigned int n);
void *alloc_private(struct lib_context *lc, const char *who, size_t size);
void *alloc_private_and_read(struct lib_context *lc, const char *who,
			     size_t size, char *path, loff_t offset);
struct raid_set *join_superset(struct lib_context *lc,
			       char *(*f_name)(struct lib_context *lc,
					       struct raid_dev *rd,
					       unsigned int subset),
			       void (*f_create)(struct raid_set *super,
					        void *private),
			       int (*f_set_sort)(struct list_head *pos,
						 struct list_head *new),
			       struct raid_set *rs,
			       struct raid_dev *rd);
int register_format_handler(struct lib_context *lc, struct dmraid_format *fmt);
int set_rs_status(struct lib_context *lc, struct raid_set *rs);
int write_metadata(struct lib_context *lc, const char *handler,
		   struct raid_dev *rd, int meta_index, int erase);

#define	to_disk	to_cpu

/* Print macros used in log methods. */

/* Undefine this to avoid offset printing in metadata logging. */
#define	NATIVE_LOG_OFFSET
#ifdef	NATIVE_LOG_OFFSET
#define	P_FMT	"0x%03x "
#define	P_OFF(x, basevar, y...) \
	((unsigned long) &x - (unsigned long) basevar), y
#else
#define	P_FMT
#define	P_OFF(x, basevar, y...)	y
#endif

#define	P(format, basevar, x, y...) \
	do { log_print(lc, P_FMT format, P_OFF(x, basevar, y)); } while(0)

#define	P2(format, basevar, i, x) \
	do { P(format, basevar, x, i, x); } while(0)
#define	DP(format, basevar, x) \
	do { P(format, basevar, x, x); } while(0)

/*
 * RAID device, set and vendor metadata retrieval macros.
 */
#define	DEVS(rs)	(!list_empty(&((struct raid_set *) (rs))->devs))
#define	SETS(rs)	(!list_empty(&((struct raid_set *) (rs))->sets))

#define	META(rd, type)	((struct type*) ((struct raid_dev*) (rd))->meta_areas[0].area)
#define	RD(pos)		(list_entry((pos), struct raid_dev, devs))
#define	RS(pos)		(list_entry((pos), struct raid_set, list))
#define RD_RS(rs)	(RD((((struct raid_set*) (rs))->devs.next)))

#define	HANDLER_LEN	sizeof(HANDLER)

#endif /* ifdef FORMAT_HANDLER */

#endif
