/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * DOS partition format handler.
 *
 * dos_read() and dos_group influenced by libparted.
 */

#define	HANDLER	"dos"

#include "internal.h"
#define	FORMAT_HANDLER
#include "dos.h"
#include <datastruct/byteorder.h>

static const char *handler = HANDLER;

/* Make up RAID device name. */
static size_t _name(struct lib_context *lc, struct raid_dev *rd,
		    unsigned short partition, char *str, size_t len,
		    unsigned char type)
{
	char part[3];

	if (type)
		snprintf(part, sizeof(part), "%u", partition);
	else
		*part = 0;

	return snprintf(str, len, "%s%s",
			get_basename(lc, rd->di->path), part) + 1;
}

static char *name(struct lib_context *lc, struct raid_dev *rd,
		   unsigned int part, unsigned char type)
{
	size_t len;
	char *ret;

	if ((ret = dbg_malloc((len = _name(lc, rd, part, NULL, 0, type)) + 1)))
		_name(lc, rd, part, ret, len, type);
	else
		log_string_err(lc, handler);

	return ret;
}

/*
 * Read a DOS partiton.
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct dos *dos)
{
	struct dos_partition *part = dos->partitions;

	for (part = dos->partitions; part < &dos->partitions[4]; part++) {
		CVT32(part->start);
		CVT32(part->length);
	}

	CVT16(dos->magic);
}
#endif

/* Allocate a DOS partition sector struct and read the data. */
static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct dos *dos, struct dev_info *di);
static struct raid_dev *dos_read(struct lib_context *lc, struct dev_info *di)
{
	struct dos *dos;
	struct dos_partition *part;
	struct raid_dev *rd = NULL;

	/* Allocate space for private ondisk metadata */
	if (!(dos = alloc_private_and_read(lc, handler, sizeof(*dos),
					   di->path, DOS_CONFIGOFFSET)))
		goto out;

	/* Check magic. */
	to_cpu(dos);
	if (dos->magic != DOS_MAGIC)
		goto bad;

	for (part = dos->partitions; part < &dos->partitions[4]; part++) {
		if (part->type == PARTITION_GPT)
			goto bad;
	}

	if (lc_opt(lc, LC_DUMP))
		log_print(lc, "%s: filing metadata not supported (use fdisk "
			  "and friends)", handler);

	/* Allocate RAID device structure (see metadata.c) */
	if (!(rd = alloc_raid_dev(lc, handler)))
		goto bad;

	/*
	 * Setup generic RAID device structure
	 * members using ondisk information.
	 */
	if (setup_rd(lc, rd, dos, di))
		goto out;

	log_err(lc, "%s: setting up RAID device %s", handler, di->path);
	free_raid_dev(lc, &rd);
	goto out;

   bad:
	dbg_free(dos);
   out:
	return rd;
}

/* Support functions for dos_group to read the partition table(s). */
static int part_is_extended (struct dos_partition *part)
{
	return part->type == PARTITION_EXT ||
	       part->type == PARTITION_EXT_LBA ||
	       part->type == PARTITION_LINUX_EXT;
}

/* Get a device offset relative to a partition. */
static uint64_t linear_start(const struct dos_partition *raw_part,
			     uint64_t offset)
{
	return (uint64_t) raw_part->start + offset;
}

/* RAID set allocation support function. */
static struct raid_set *_alloc_raid_set(struct lib_context *lc,
					struct raid_dev *rd)
{
	struct raid_set *rs;

	if ((rs = find_set(lc, NULL, rd->name, FIND_TOP)))
		LOG_ERR(lc, NULL, "%s: RAID set %s already exists",
			handler, rs->name);

	if (!(rs = alloc_raid_set(lc, handler)))
		return NULL;

	rs->status = rd->status;
	rs->type   = rd->type;

       	if (!(rs->name = dbg_strdup(rd->name))) {
		dbg_free(rs);
		rs = NULL;
		log_string_err(lc, handler);
	}

	return rs;
}

/*
 * Allocate a DOS RAID device and a set.
 * Set the device up and add it to the set.
 */
static int _create_rs_and_rd(struct lib_context *lc, struct raid_dev *rd,
			     struct dos_partition *raw_part, uint64_t sector,
			     unsigned int part)
{
	struct raid_dev *r;
	struct raid_set *rs;

	if (!(r = alloc_raid_dev(lc, handler)))
		return 0;

	if (!(r->di = alloc_dev_info(lc, rd->di->path)))
		goto free_raid_dev;

        if (!(r->name = name(lc, rd, part, 1)))
		goto free_di;

	r->fmt = rd->fmt;
	r->status = rd->status;
	r->type = rd->type;

	if ((uint64_t) raw_part->start > sector)
		sector = 0;

	r->offset = linear_start(raw_part, sector);
	r->sectors = (uint64_t) raw_part->length;

	if (!(rs = _alloc_raid_set(lc, r)))
		goto free_di;

	list_add_tail(&r->devs, &rs->devs);
	list_add_tail(&rs->list, LC_RS(lc));

	return 1;

  free_di:
	free_dev_info(lc, r->di);
  free_raid_dev:
	free_raid_dev(lc, &r);

	return 0;
}

/*
 * RAID set grouping.
 *
 * In this case it is not really about grouping, more about providing
 * the propper input to the activation layer by creating a RAID set per
 * logical partition and a RAID device hanging off it mapping the partition
 * linearly.
 *
 * Partition code inspired by libparted and squeezed for this purpose.
 */
struct out_of_order {
	uint64_t lo2;
	int out_of_order;
};

static int is_partition(struct dos_partition *p, struct out_of_order *ooo,
			uint64_t start_sector)
{
	return p->type != PARTITION_EMPTY && p->length &&
	       (start_sector ? ooo->out_of_order ||
			       ((uint64_t) p->start < start_sector) : 1);
}

/* FIXME: skip combined/deleted/recreated partitions. */
static uint64_t next_partition(uint64_t start_sector,
			       struct out_of_order *ooo,
			       struct dos_partition *p1,
			       struct dos_partition *p2)
{
	uint64_t o1, o2, o3, ret;

	o1 = linear_start(p1, start_sector + (uint64_t) p1->length);
	o2 = linear_start(p2, (uint64_t) p2->length);
	o3 = start_sector + (uint64_t) p2->length;

	if (ooo->out_of_order && start_sector == ooo->lo2)
		ret = start_sector;
	else if ((uint64_t) p1->start > start_sector)
		ret = o3;
	else if (o2 < start_sector)
		ret = o2;
	else if (o1 == o3)
		ret = o2;
	else
		ret = o1;

	ooo->lo2 = o2;
	if (ret < start_sector)
		ooo->out_of_order = 1;

	return ret;
}

static int group_rd_extended(struct lib_context *lc, struct raid_dev *rd,
			     struct out_of_order *ooo,
			     uint64_t start_sector, unsigned int part)
{
	int ret = 0;
	uint64_t new_start_sector;
	struct dos *dos;
	struct dos_partition *p1, *p2;

	/* Allocate and read an extended partition table. */
	if (!(dos = alloc_private_and_read(lc, handler, sizeof(*dos),
					   rd->di->path, start_sector << 9)))
		return 0;

	/* Weird: empty extended partitions are filled with 0xF6 by PM */
	to_cpu(dos);
	if (dos->magic == PARTITION_MAGIC_MAGIC)
		goto out;

	p1 = &dos->partitions[0];
	p2 = &dos->partitions[1];

	if (is_partition(p1, ooo, start_sector) &&
	    !_create_rs_and_rd(lc, rd, p1, start_sector, part++))
		goto out;

	/* Recurse into the partition chain. */
	if ((new_start_sector = next_partition(start_sector, ooo, p1, p2))) {
		if (new_start_sector != start_sector &&
	    	    !group_rd_extended(lc, rd, ooo, new_start_sector, part))
			goto out;
	}

	ret = 1;

  out:
	dbg_free(dos);
	return ret;
}

/* Handle primary partitions. */
static int group_rd(struct lib_context *lc, struct raid_dev *rd,
		    uint64_t start_sector)
{
	unsigned int i;
	uint64_t part_start, extended_part_start = 0;
	struct dos *dos = META(rd, dos);
	struct dos_partition *raw_part;
	struct out_of_order ooo = { 0, 0 };

	/*
	 * Walk 4 array entries of a primary partition table.
	 *
	 * FIXME: check integrity of partition's C/H/S values.
	 */
	for (i = 0; i < 4; i++) {
		raw_part = &dos->partitions[i];
		if (!is_partition(raw_part, &ooo, start_sector))
			continue;

		part_start = linear_start(raw_part, start_sector);

		/* Avoid infinite recursion. */
		if (part_start == start_sector ||
		    part_start > rd->sectors)
			continue;

		if (part_is_extended(raw_part))
			extended_part_start = part_start;
		else if (!_create_rs_and_rd(lc, rd, raw_part,
					    start_sector, i + 1))
			return 0;
	}

	return extended_part_start ?
	       group_rd_extended(lc, rd, &ooo, extended_part_start, 5) : 1;
}

/* Add a DOS RAID device to a set */
static struct raid_set *dos_group(struct lib_context *lc,
				    struct raid_dev *rd)
{
	/*
	 * Once we get here, a DOS boot sector with a partition table
	 * has been discovered by dos_read.
	 *
	 * We need to run through the (nested) partition tables and create
	 * a RAID set and a linear RAID device hanging off it for every primary
	 * and logical partition, so that the activate code is happy.
	 */
	return group_rd(lc, rd, 0) ? (struct raid_set*) 1 : NULL;
}

/*
 * Check integrity of a DOS RAID set.
 */
static int dos_check(struct lib_context *lc, struct raid_set *rs)
{
	return 1; /* Nice, eh ? */
}

static struct dmraid_format dos_format = {
	.name	= HANDLER,
	.descr	= "(+) DOS partitions on SW RAIDs",
	.format = FMT_PARTITION,
	.read	= dos_read,
	.write	= NULL, /* Not supported */
	.group	= dos_group,
	.check	= dos_check,
};

/* Register this format handler with the format core */
int register_dos(struct lib_context *lc)
{
	return register_format_handler(lc, &dos_format);
}

/*
 * Set the RAID device contents up derived from the DOS ones
 *
 * For a DOS partition we essentially just save the
 * partition table sector and let dos_group do the rest...
 */
static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct dos *dos, struct dev_info *di)
{
	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas->offset = DOS_CONFIGOFFSET >> 9;
	rd->meta_areas->size = sizeof(*dos);
	rd->meta_areas->area = (void*) dos;

        rd->di  = di;
	rd->fmt = &dos_format;

	rd->offset   = DOS_DATAOFFSET;
	rd->sectors  = di->sectors;
	rd->sectors -= rd->sectors & 1;

        rd->status = s_ok; /* Always :) */
	rd->type   = t_partition;

        if ((rd->name = name(lc, rd, 0, 0)))
		return 1;

	return 0;
}
