/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include <errno.h>
#include <sys/file.h>
#include "internal.h"
#include "locking.h"

/* File locking private data. */
static const char *lock_file = "/var/lock/dmraid/.lock";
static int lf = -1;

/* flock file. */
static int lock(struct lib_context *lc, struct resource *res)
{
	int ret = 0;

	/* Already locked. */
	if (lf > -1)
		return 1;

	log_warn(lc, "locking %s", lock_file);
	if ((lf = open(lock_file, O_CREAT | O_APPEND | O_RDWR, 0777)) < 0) {
		log_err(lc, "opening lockfile %s", lock_file);
		goto out;
	}

	if (flock(lf, LOCK_EX)) {
		log_err(lc, "flock lockfile %s", lock_file);
		close(lf);
		lf = -1;
		goto out;
	}

	ret = 1;

  out:
	return ret;
}

/* Unlock file. */
static void unlock(struct lib_context *lc, struct resource *res)
{
	if (lf == -1)
		return;

	log_warn(lc, "unlocking %s", lock_file);
	unlink(lock_file);
	if (flock(lf, LOCK_NB | LOCK_UN))
		log_err(lc, "flock lockfile %s", lock_file);

	if (close(lf))
		log_err(lc, "close lockfile %s", lock_file);

	lf = -1;
}

/* File base locking interface. */
static struct locking file_locking = {
	.name	 = "file",
	.lock 	 = lock,
	.unlock	 = unlock,
};

static int init_file_locking(struct lib_context *lc)
{
	int ret = 0;
	char *dir;

	if (!(dir = get_dirname(lc, (char*) lock_file)))
		return 0;

	if (!mk_dir(lc, dir))
		goto out;

	/* Fail on read-only file system. */
	if (access(dir, R_OK | W_OK) && errno == EROFS)
		goto out;

	lc->lock = &file_locking;
	ret = 1;

   out:
	dbg_free(dir);

	return ret;
}

/*
 * External locking interface.
 */

/* Initialize locking. */
int init_locking(struct lib_context *lc)
{
	if (lc->locking_by_name)
		BUG(lc, 0, "no locking selection yet");

	return init_file_locking(lc);
}

/* Hide locking. */
int lock_resource(struct lib_context *lc, struct resource *res)
{
	return lc->lock->lock(lc, res);
}

/* Hide unlocking. */
void unlock_resource(struct lib_context *lc, struct resource *res)
{
	 lc->lock->unlock(lc, res);
}
