/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include <errno.h>
#include "internal.h"

/* Create directory recusively. */
static int mk_dir_recursive(struct lib_context *lc, const char *dir)
{
        char *orig, *s;

        log_notice(lc, "creating directory %s", dir);

        /* Create parent directories */
        orig = s = dbg_strdup((char*) dir);
	while (s) {
		if ((s = strchr(s + 1, '/')))
			*s = '\0';

		if (mkdir(orig, 0777) && errno != EEXIST) {
			log_err(lc, "mkdir %s", orig);
			dbg_free(orig);
			return 0;
		}

		if (s)
			*s++ = '/';
	}

	dbg_free(orig);

        return 1;
}

/* Create directory. */
int mk_dir(struct lib_context *lc, const char *dir)
{
	struct stat info;

	if (stat(dir, &info))
		return mk_dir_recursive(lc, dir);

	if (S_ISDIR(info.st_mode))
		return 1;

	LOG_ERR(lc, 0, "directory %s not found", dir);
}

static int rw_file(struct lib_context *lc, const char *who, int flags,
		   char *path, void *buffer, size_t size, loff_t offset)
{
	int fd, ret = 0;
	loff_t o;

	if ((fd = open(path, flags, lc->mode)) == -1)
		LOG_ERR(lc, 0, "opening \"%s\"\n", path);

	if (offset && (o = lseek64(fd, offset, SEEK_SET)) == (loff_t) -1) {
		log_err(lc, "%s: seeking device \"%s\" to %" PRIu64,
			who, path, offset);
		goto err;
	}

	if ((ret = ((flags & O_WRONLY) ? write(fd, buffer, size) :
					 read(fd, buffer, size)) != size)) {
		log_err(lc, "%s: %sing %s[%d]", who,
			(flags & O_WRONLY) ? "writ" : "read", path, errno);
		ret = 0;
		goto err;
	}

	ret = 1;

  err:
	close(fd);

	return ret;
}

int read_file(struct lib_context *lc, const char *who, char *path,
	      void *buffer, size_t size, loff_t offset)
{
	return rw_file(lc, who, O_RDONLY, path, buffer, size, offset);
}

int write_file(struct lib_context *lc, const char *who, char *path,
	       void *buffer, size_t size, loff_t offset)
{
	/* O_CREAT|O_TRUNC are noops on a devnode. */
	return rw_file(lc, who, O_WRONLY|O_CREAT|O_TRUNC, path,
		       buffer, size, offset);
}
