/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include <stdarg.h>
#include "internal.h"

/* Prompt for a yes/no answer */
int yes_no_prompt(struct lib_context *lc, const char *prompt, ...)
{
	int c = 0;
	va_list ap;

	while (c != 'y' && c != 'n') {
		if (c == '\n' || !c) {
			va_start(ap, prompt);
			vprintf(prompt, ap);
			va_end(ap);
			printf(" ? [y/n] :");
		}
		c = tolower(getchar());
	}

	while (getchar() != '\n');

	return c == 'y';
}

/* Remove any whitespace from a string. */
char *remove_space(struct lib_context *lc, char *str, size_t size)
{
	unsigned int end = size;
	char *ws = str;

	while ((ws = strpbrk(ws, " \t\n"))) {
		if (ws > &str[size-2]) {
			if (ws == &str[size-1])
				str[size-1] = 0;

			break;
		}

		memmove(ws, ws + 1, size - (ws - str));
		str[--end] = 0;
	}

	return str;
}

/* Return the basename of a path. */
char *get_basename(struct lib_context *lc, char *str)
{
	char *ret = strrchr(str, '/');

	return ret ? ret + 1 : str;
}

/* Return the dirname of a path. */
char *get_dirname(struct lib_context *lc, char *str)
{
	size_t len;
	char *ret;

	len = (ret = strrchr(str, '/')) ? ret - str : strlen(str);
	if ((ret = dbg_malloc(len + 1)))
		strncpy(ret, str, len);
		
	return ret;
}

/* Convert a numeric string to alpha. */
void mk_alpha(struct lib_context *lc, char *str, size_t len)
{
	char buf[2];
	static const char asc[] = "abcdefghij";

	for (buf[1] = 0; *str && len; str++, len--) {
		if (isdigit(*str)) {
			buf[0] = *str;
			*str = asc[atoi(buf)];
		}
	}
}

/* Grow a string. */
static int grow_string(struct lib_context *lc, char **string, const char *s)
{
	size_t len;
	char *tmp = *string;

 	len = strlen(s) + (tmp ? strlen(tmp) + 1 : 1);
	if ((*string = dbg_realloc(*string, len))) {
		if (!tmp)
			**string = '\0';
	} else if (tmp)
		dbg_free(tmp);
	
	return *string ? 1 : 0;
}

/* Free a string. */
void free_string(struct lib_context *lc, char **string)
{
	if (*string) {
		dbg_free(*string);
		*string = NULL;
	}
}

/* Push a string onto the end of another. */
static int p_str(struct lib_context *lc, char **string, const char *s)
{
	int ret;

	if ((ret = grow_string(lc, string, s)))
		strcat (*string, s);

	return ret;
}

/* Push a string defined by a start and end pointer onto the end of another. */
static int p_str_str(struct lib_context *lc, char **string,
		     char *begin, char *end)
{
	int ret = 1;

	if (end != begin) {
		*end = 0;
		ret = p_str(lc, string, begin);
	}

	return ret;
}

/* Push an uint64_t in ascii onto the end of a string. */
static int p_u64(struct lib_context *lc, char **string, const uint64_t u)
{
	char buffer[22];

	sprintf(buffer, "%" PRIu64, u);

	return p_str(lc, string, buffer);
}

/* Push an uint_t in ascii onto the end of a string. */
static int p_u(struct lib_context *lc, char **string, const unsigned int u)
{
	return p_u64(lc, string, (uint64_t) u);
}

static char *find_percent(char *s)
{
	for (; *s; s++) {
		if (*s == '%')
			return s;
	}

	return NULL;
}

/* Push a format string defined list of arguments onto a string. */
int p_fmt(struct lib_context *lc, char **string, const char *fmt, ...)
{
	int ret = 1;
	char *b, *f, *f_sav;
	va_list ap;
	
	if (!(f = f_sav = dbg_strdup((char *) fmt)))
		return 0;

	va_start(ap, fmt);
	while (ret && *(b = f++)) {
		if (!(f = find_percent(b))) {
			/* No '%' -> just print string. */
			ret = p_str(lc, string, b);
			break;
		}

		if (!(ret = p_str_str(lc, string, b, f)))
			break;

		switch (*++f) {
		case 's':
			ret = p_str(lc, string, va_arg(ap, char *));
			break;
	
		case 'u':
			ret = p_u(lc, string, va_arg(ap, unsigned int));
			break;

		case 'U':
			ret = p_u64(lc, string, va_arg(ap, uint64_t));
			break;

		default:
			ret = 0;
		}

		f++;
	}

	va_end(ap);
	dbg_free(f_sav);

	return ret;
}
