/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Tool library
 */

#include <dmraid.h>
#include <ctype.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include "commands.h"
#include "toollib.h"

/* Toolib init. */
int toollib_init(struct lib_context *lc)
{
	format_names = NULL;

	return 1;
}

/* Toolib cleanup. */
void toollib_exit(struct lib_context *lc)
{
	free_format_names(lc);
}


/* [De]activate a RAID set. */
static int _change_set(struct lib_context *lc, void *rs, int arg)
{
	int ret;

	/* FIXME: startup monitoring RAID set on activate; stop on deactivate */
	ret = (ACTIVATE & action) ? activate_set(lc, rs) :
				    deactivate_set(lc, rs);

	if (ret)
		log_info(lc, "%sctivated %s RAID set \"%s\"",
			 action & ACTIVATE ? "A": "Dea",
			 get_set_type(lc, rs),
			 get_set_name(lc, rs));

	return ret;
}

/* [De]activate RAID sets. */
int change_sets(struct lib_context *lc, int arg)
{
	/* Discover partitions to deactivate RAID sets for and work on them. */
	if (DEACTIVATE & action) {
		discover_partitions(lc);
		process_sets(lc, _change_set, 0, PARTITIONS);
	}

	process_sets(lc, _change_set, arg, SETS);

	/* Discover partitions to activate RAID sets for and work on them. */
	if ((ACTIVATE & action) && !(NO_PARTITIONS & action)) {
		discover_partitions(lc);
		process_sets(lc, _change_set, 0, PARTITIONS);
	}

	return 1;
}

/* Build all sets or the ones given. */
void build_sets(struct lib_context *lc, char **sets)
{
	int o = 0;

	do {
		if (!group_set(lc, sets[o]))
			log_err(lc, "building set");

		if (!sets[o])
			break;
	} while (sets[++o]);
}

/* Convert a character sting to lower case. */
void str_tolower(char *s)
{
	for (; *s; s++)
		*s = tolower(*s);
}

/* Check if selected or all formats shall be used to read the metadata. */
void free_format_names(struct lib_context *lc)
{
	if (format_names) {
		unsigned int i = 0;

		for (i = 0; format_names[i]; i++)
			dbg_free(format_names[i]);

		dbg_free(format_names);
		format_names = NULL;
	}
}

static int push_format_name(struct lib_context *lc, char ***fmt_names,
			    char *name)
{
	unsigned int i;
	char **p = *fmt_names;

	for (i = 0; p && p[i]; i++);
	if (!(p = dbg_realloc(p, (i + 2) * sizeof (*p)))) {
		free_format_names(lc);
		return 0;
	}

	p[i] = name;
	p[i+1] = NULL;
	*fmt_names = p;

	return 1;
}

int valid_format(struct lib_context *lc, char *fmt,
		 char ***fmt_names, char **invalid)
{
	int ret = 1;
	char *name, *p = remove_space(lc, fmt, strlen(fmt)), *sep;

	while (*p && ret) {
		if ((sep = strchr(p, ',')))
			*sep = '\0';

		if (!(name = dbg_strdup(p)))
			return 0;

		if (!(ret = check_valid_format(lc, name)))
			*invalid = name;

		/*
		 * Even push an invalid format name,
		 * so that it will be freed on cleanup.
		 */
		if (!push_format_name(lc, fmt_names, name))
			return 0;

		p += strlen(p);
		if (sep) {
			*sep = ',';
			p++;
		}
	}

	return ret;
}

void format_error(struct lib_context *lc, char **fmt_names, const char *error)
{
	if (fmt_names) {
		unsigned int i;

		log_print_nnl(lc, "No RAID %s with format: ", error);
		for (i = 0; fmt_names[i]; i++) {
			log_print_nnl(lc, "%s", fmt_names[i]);
			if (fmt_names[i+1])
				log_print_nnl(lc, ",");
		}
		log_print_nnl(lc, "\n");
	} else
		log_print(lc, "No Software RAID %ss", error);
}
